/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang, 
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


#include "node_list_simple.h"
#include <assert.h>
#include <stdio.h>

node_list_simple::node_list_simple( wfs_node me ) {

  _me = me; // TODO: move to a parent constructor.  I am lazy.
  _nodes = new nls_node();
  _nodes->node = me;
  _nodes->prev = NULL;
  _nodes->next = NULL;

  assert( pthread_mutex_init( &_nodes_mutex, NULL ) == 0 );

}

node_list_simple::~node_list_simple() {

  pthread_mutex_lock( &_nodes_mutex );
  while( _nodes != NULL ) {
    nls_node *x = _nodes;
    _nodes = _nodes->next;
    delete x;
  }
  pthread_mutex_unlock( &_nodes_mutex );
  pthread_mutex_destroy( &_nodes_mutex );

}

// lock should be held by the caller
nls_node *node_list_simple::successor_node( wfs_node_id node_id, int n ) {
  // go through the list as O(n), and just find the first node that
  // doesn't exceed the id.
  nls_node *answer = NULL;
  nls_node *curr_node = _nodes;
  nls_node *first_succ = NULL;
  int succ_num = 0;
  bool looped = false;

  while( curr_node != NULL ) {

    if( curr_node->node.id >= node_id || looped ) {

      if( succ_num == 0 ) {
	first_succ = curr_node;
      } else if( first_succ->node.id == curr_node->node.id ) {
	// If there are N < n nodes in the ring, this should ensure
	// that the Nth successor is returned
	break;
      }

      succ_num++;
      answer = curr_node;
      if( succ_num >= n ) {
	break;
      }
    }
    
    curr_node = curr_node->next;

    if( curr_node == NULL && !looped ) {
      curr_node = _nodes;
      looped = true;
    }

  }

  return answer;

}

wfs_node node_list_simple::successor( wfs_node_id node_id, int n ) {

  pthread_mutex_lock( &_nodes_mutex );
  wfs_node answer = successor_node( node_id, n )->node;
  pthread_mutex_unlock( &_nodes_mutex );
  return answer;

}


wfs_node node_list_simple::predecessor( wfs_node_id node_id ) {

  pthread_mutex_lock( &_nodes_mutex );
  nls_node *answer = successor_node( node_id );
  if( answer->prev == NULL ) {
    // return the last node
    answer = _nodes;
    while( answer && answer->next != NULL ) {
      answer = answer->next;
    }
  } else {
    answer = answer->prev;
  }
  pthread_mutex_unlock( &_nodes_mutex );
  return answer->node;

}

void node_list_simple::sync_to_node_list( std::vector<wfs_node> nodes ) {

  // TODO: inefficient, as always
  
  // first clear the existing list
  pthread_mutex_lock( &_nodes_mutex );
  while( _nodes != NULL ) {
    nls_node *x = _nodes;
    _nodes = _nodes->next;
    delete x;
  }

  _nodes = new nls_node();
  _nodes->node = _me;
  _nodes->prev = NULL;
  _nodes->next = NULL;

  // now add all the new nodes
  for( unsigned i = 0; i < nodes.size(); i++ ) {
    add_node_locked( nodes[i] );
  }

  pthread_mutex_unlock( &_nodes_mutex );

}

void node_list_simple::add_node_locked( wfs_node node ) {

  nls_node *new_node = new nls_node();
  new_node->node = node;
  new_node->prev = NULL;
  new_node->next = NULL;

  nls_node *curr_node = _nodes;
  nls_node *last_node = NULL;
  bool inserted = false;
  while( curr_node != NULL ) {

    if( curr_node->node.id == node.id ) {
      curr_node->node.site_id = node.site_id;
      delete new_node;
      return; // no dups
    } else if( curr_node->node.id > node.id ) {

      new_node->prev = curr_node->prev;
      new_node->next = curr_node;
      curr_node->prev = new_node;
      if( new_node->prev != NULL ) {
	new_node->prev->next = new_node;
      } else {
	_nodes = new_node;
      }
      inserted = true;
      break;

    }

    last_node = curr_node;
    curr_node = curr_node->next;

  }

  if( !inserted ) {
    last_node->next = new_node;
    new_node->prev = last_node;
  }

}

void node_list_simple::add_node( wfs_node node ) {

  
  pthread_mutex_lock( &_nodes_mutex );
  add_node_locked( node );
  pthread_mutex_unlock( &_nodes_mutex );

}

wfs_node node_list_simple::get_node( int index ) {

  pthread_mutex_lock( &_nodes_mutex );

  // TODO: this is super slow
  nls_node *curr_node = _nodes;
  int n = 0;
  while( n < index && curr_node != NULL ) {
    curr_node = curr_node->next;
    n++;
  }

  pthread_mutex_unlock( &_nodes_mutex );

  if( curr_node == NULL ) {
    wfs_node x;
    x.id = 0;
    return x;
  } else {
    return curr_node->node;
  }

}
