/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang, 
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


#ifndef __REPLICATION_POLICY_H
#define __REPLICATION_POLICY_H 1

#include <iostream>
#include "wfs_slice.h"
#include "host.h"

struct rep_policy {
  /*
   * placement gives hints as to which parameters the manifestation should satisfy.
   * rep_num must ALWAYS be satisfied, but we may not care about about site placement.
   * It makes no sense to force a default policy in those cases.
   */
  enum site_placement {
    ANY,	// Any placement is fine. Ignore both site_num and prim_site.
    LOCAL,	// Ignore only site_num. Must satisfy prim_site.
    STRICT	// Satisfy all fields.
  };
  explicit rep_policy(unsigned int repnum = 1, rep_policy::site_placement pol = rep_policy::STRICT ,std::string psite = "",unsigned int sitenum = 1 )
  : rep_num(repnum), site_pol(pol), prim_site(psite), site_num(sitenum) {};

  unsigned int rep_num;		// How many replicas
  site_placement site_pol;	// The site placement policy
  std::string prim_site;	// Site of primary server
  unsigned int site_num;	// Number of different sites to keep replicas on

  operator std::string();
};

/*
 * slice_assignment is a manifestation/implementation of a policy, represented as a list of replicas
 * (The first being the primary). It also keeps the policy itself.
 * In its opaque string state it represents the state of a lock.
 */
struct slice_assignment {
  slice_assignment(std::string, unsigned int exp = 0);
  slice_assignment(wfs_slice sl, rep_policy pol, std::list<host> lst, unsigned int exp = 0)
  : slice(sl) , policy(pol), replicas(lst), lease_period(exp) {};
  /*
   * Transforms the slice_assignment into an opaque string which will be used as the lock_state when dealing with
   * the configuration server.
   */
  const std::string to_canonical_string(void) const;
  /*
   * Rebuild this assignment from a state string
   */
  void from_canonical_string(const std::string state);
  /*
   * Check the replica list against the policy.
   */
  bool validate(std::string& err) const;
  const host get_primary() const {return replicas.front();};

  wfs_slice slice;
  rep_policy policy;
  std::list<host> replicas;
  unsigned int lease_period;  // Time to expiration in seconds
};

std::ostream& operator<<(std::ostream&, const rep_policy&);
std::istream& operator>> (std::istream&, rep_policy&);
std::ostream& operator<<(std::ostream&, const slice_assignment&);
std::istream& operator>> (std::istream&, slice_assignment&);

#endif // __REPLICATION_POLICY_H
