/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang, 
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


// replicated state machine interface.

#ifndef rsm_h
#define rsm_h

#include <string>
#include <vector>
#include <list>
#include "rpc.h"
#include "rsm_protocol.h"
#include "rsm_state_transfer.h"
#include "rsm_log.h"
#include <assert.h>
#include <arpa/inet.h>
#include "paxos.h"
#include "host_marshall.h"

struct arg_t {
  arg_t (unsigned _seqno, int p, std::string _req) {
    seqno = _seqno;
    procno = p;
    req = _req;
  }
  unsigned seqno;
  int procno;
  std::string req;
};

class rsm_reserved {
 public:
  enum rpc_numbers {
    became_master = 0xffff,
    lost_master = 0xfffe
  };
};

class rsm_handler;

class rsm : public paxos_view_change {
 private:
  void reg1(int proc, rsm_handler *);
 protected:
  std::map<int, rsm_handler *> procs;
  paxos *pxs;
  class rsm_state_transfer *stf;
  std::map<std::string, rpcc *> cl;
  rpcs *rsmrpc;
  viewstamp myvs;
  viewstamp last_myvs;
  std::string master;
  //if stable is false, this node's rsm state
  //might not be up-to-date due to view changes
  bool stable;

  int ops_since_last_cp;

  std::list<arg_t*> requests;
  void insert(arg_t *a);

  rsm_log *rlog;

  rsm_client_protocol::status client_members(int i,
               std::vector<std::string> &r);
  rsm_protocol::status invoke(std::vector<rsm_record> recs, int &dummy);
  rsm_protocol::status transferreq(std::string src, viewstamp vs,
           std::string &r);
  rsm_protocol::status cptransferreq(std::string src, viewstamp fromvs, rsm_record &r);
  rsm_protocol::status logtransferreq(std::string src, viewstamp fromvs, viewstamp tovs,
           std::vector<rsm_record> &r);
  rsm_protocol::status viewstampreq(unsigned int vid, viewstamp &r);

  pthread_mutex_t rsm_mutex;
  pthread_cond_t recovery_cond;

  std::string execute(int procno, std::string req, bool replicated = true);
  bool statetransfer(std::string m, viewstamp vs);
  bool logtransfer(std::string m, viewstamp vs);
  std::string get_master();
  bool find_master(viewstamp &vs);

  struct invoke_rec {
    static const int NOTDONE = -10;
  invoke_rec(int p, std::string r) : procno(p), req(r),
      ret(NOTDONE), next(NULL) {}
    int procno;
    std::string req;
    int ret;
    std::string r;
    invoke_rec *next;
  };
  struct invoke_rec *invoke_rec_head;
  struct invoke_rec *invoke_rec_tail;
  pthread_mutex_t invoke_rec_mutex;

 public:
  rsm (std::string _first, std::string _me, std::string log_dir,
       wfs_client_manager *clients, int pxs_delay_max = 10);
  ~rsm();

  enum dec_ret {
    NO = 0, // Decide no.
    YES_REP = 20,  // Decide yes. Need to replicate
    YES_NOREP = 21 // Decide yes. Don't replicate. (like a read only operation.)
  };

  rsm_client_protocol::status client_invoke(int procno, std::string req,
					    std::string &r);

  bool amimaster();
  // Calling from within handlers that are already under rsm_mutex locked.
  bool amimaster_mx();
  std::string myaddr() { return pxs->myaddr(); };
  void set_state_transfer(rsm_state_transfer *_stf) { stf = _stf; };
  void recovery();
  void recover_from_log();
  void view_change();
  viewstamp get_latest_viewstamp() const {return last_myvs;};
  viewstamp get_viewstamp() const {return myvs;};
  void call_invoke(std::string mem, std::vector<rsm_record> recs);

  template<class S, class A1, class R>
    void reg(int proc, S*, int (S::*meth)(const A1 a1, R &)
    	,dec_ret (S::*dec)(const A1 a1, R &, int &) = NULL);
  template<class S, class A1, class A2, class R>
    void reg(int proc, S*, int (S::*meth)(const A1 a1, const A2 a2, R &)
    ,dec_ret (S::*dec)(const A1 a1, const A2 a2, R &, int &) = NULL);
  template<class S, class A1, class A2, class A3, class R>
    void reg(int proc, S*, int (S::*meth)(const A1 a1, const A2 a2,
            const A3 a3, R &),dec_ret (S::*dec)(const A1 a1, const A2 a2,const A3 a3, R &, int &) = NULL);
  template<class S, class A1, class A2, class A3, class A4, class R>
    void reg(int proc, S*, int (S::*meth)(const A1 a1, const A2 a2,
            const A3 a3, const A4 a4, R &)
            ,dec_ret (S::*dec)(const A1 a1, const A2 a2,const A3 a3, const A4 a4, R &, int &) = NULL);
  template<class S, class A1, class A2, class A3, class A4, class A5, class R>
    void reg(int proc, S*, int (S::*meth)(const A1 a1, const A2 a2,
            const A3 a3, const A4 a4,const A5 a5, R &)
            ,dec_ret (S::*dec)(const A1 a1, const A2 a2,const A3 a3, const A4 a4,const A5 a5, R &, int &) = NULL);

  // Read Only Handlers.
  template<class S, class A1, class R>
    void reg_ro(int proc, S* s, int (S::*meth)(const A1 a1, R &)){ rsmrpc->reg(proc,s,meth);}
  template<class S, class A1, class A2, class R>
    void reg_ro(int proc, S* s, int (S::*meth)(const A1 a1, const A2 a2, R &)){ rsmrpc->reg(proc,s,meth);}
  template<class S, class A1, class A2, class A3, class R>
    void reg_ro(int proc, S* s, int (S::*meth)(const A1 a1, const A2 a2,
            const A3 a3, R &)){ rsmrpc->reg(proc,s,meth);}
  template<class S, class A1, class A2, class A3, class A4, class R>
    void reg_ro(int proc, S* s, int (S::*meth)(const A1 a1, const A2 a2,
            const A3 a3, const A4 a4, R &)){ rsmrpc->reg(proc,s,meth);}
  template<class S, class A1, class A2, class A3, class A4, class A5, class R>
    void reg_ro(int proc, S* s, int (S::*meth)(const A1 a1, const A2 a2,
            const A3 a3, const A4 a4,
            const A5 a5, R &)){ rsmrpc->reg(proc,s,meth);}
};

// Note! The down side for this implementation is that it does a redundant marshall/unmarshall
// and duplicates code in reg. I don't like pushing the call to inside fn,
// since we can't assume that decision functions are the norm, so  keep the common case simple
// and the 2 methods independent. Reconsider this? Y.S.
//class decision_rsm_handler : public rsm_handler  {
//  public:
//      virtual bool decide(unmarshall &, marshall &) = 0;
//};
class rsm_handler {
    public:
        rsm_handler();
        virtual ~rsm_handler() { }
        virtual int fn(unmarshall &, marshall &) = 0;
        // Since the decision function returns a boolean we need an extra placeholder for the application's int reply.
        virtual rsm::dec_ret decide(unmarshall &, marshall &, int &) {return rsm::YES_REP;};
        bool getusedecision() { return _usedecision;}
        void setusedecision(bool b) { _usedecision = b;}
    private:
        bool _usedecision;
};

template<class S, class A1, class R> void
  rsm::reg(int proc, S*sob, int (S::*meth)(const A1 a1, R & r)
  ,dec_ret (S::*dec)(const A1 a1, R & r, int & r2))
{
  class h1 : public rsm_handler {
  private:
    S * sob;
    int (S::*meth)(const A1 a1, R & r);
    dec_ret (S::*dec)(const A1 a1, R & r, int & r2);
  public:
  h1(S *xsob, int (S::*xmeth)(const A1 a1, R & r),dec_ret (S::*xdec)(const A1 a1,R & r, int & r2))
      : sob(xsob), meth(xmeth), dec(xdec) { }
    int fn(unmarshall &args, marshall &ret) {
      A1 a1;
      R r;
      args >> a1;
      assert(args.okdone());
      int b = (sob->*meth)(a1,r);
      ret << r;
      return b;
    }
    dec_ret decide(unmarshall &args, marshall &ret, int &wccret) {
      if(!dec) return YES_REP;
      A1 a1;
      R r;
      args >> a1;
      assert(args.okdone());
      dec_ret b = (sob->*dec)(a1,r,wccret);
      ret << r;
      return b;
    }
  };
  rsm_handler * hp =  new h1(sob, meth, dec);
  if(dec) hp->setusedecision(true);
  reg1(proc,hp);
}

template<class S, class A1, class A2, class R> void
  rsm::reg(int proc, S*sob, int (S::*meth)(const A1 a1, const A2 a2, R & r)
  ,dec_ret (S::*dec)(const A1 a1, const A2 a2, R & r, int & r2))
{
 class h1 : public rsm_handler {
  private:
    S * sob;
    int (S::*meth)(const A1 a1, const A2 a2, R & r);
    dec_ret (S::*dec)(const A1 a1, const A2 a2, R & r, int & r2);
  public:
  h1(S *xsob, int (S::*xmeth)(const A1 a1, const A2 a2, R & r),dec_ret (S::*xdec)(const A1 a1, const A2 a2,R & r, int & r2))
    : sob(xsob), meth(xmeth), dec(xdec) { }
    int fn(unmarshall &args, marshall &ret) {
      A1 a1;
      A2 a2;
      R r;
      args >> a1;
      args >> a2;
      assert(args.okdone());
      int b = (sob->*meth)(a1,a2,r);
      ret << r;
      return b;
    }
    dec_ret decide(unmarshall &args, marshall &ret, int &wccret) {
      if(!dec) return YES_REP;
      A1 a1;
      A2 a2;
      R r;
      args >> a1;
      args >> a2;
      assert(args.okdone());
      dec_ret b = (sob->*dec)(a1,a2,r,wccret);
      ret << r;
      return b;
    }
  };
  rsm_handler * hp =  new h1(sob, meth, dec);
  if(dec) hp->setusedecision(true);
  reg1(proc,hp);
}

template<class S, class A1, class A2, class A3, class R> void
  rsm::reg(int proc, S*sob, int (S::*meth)(const A1 a1, const A2 a2,
             const A3 a3, R & r)
             ,dec_ret (S::*dec)(const A1 a1, const A2 a2, const A3 a3, R & r, int & r2))
{
 class h1 : public rsm_handler {
  private:
    S * sob;
    int (S::*meth)(const A1 a1, const A2 a2, const A3 a3, R & r);
    dec_ret (S::*dec)(const A1 a1, const A2 a2, const A3 a3, R & r, int & r2);
  public:
  h1(S *xsob, int (S::*xmeth)(const A1 a1, const A2 a2, const A3 a3, R & r),dec_ret (S::*xdec)(const A1 a1, const A2 a2, const A3 a3,
                    R & r, int & r2))
    : sob(xsob), meth(xmeth), dec(xdec) { }
    int fn(unmarshall &args, marshall &ret) {
      A1 a1;
      A2 a2;
      A3 a3;
      R r;
      args >> a1;
      args >> a2;
      args >> a3;
      assert(args.okdone());
      int b = (sob->*meth)(a1,a2,a3,r);
      ret << r;
      return b;
    }
    dec_ret decide(unmarshall &args, marshall &ret, int &wccret) {
      if(!dec) return YES_REP;
      A1 a1;
      A2 a2;
      A3 a3;
      R r;
      args >> a1;
      args >> a2;
      args >> a3;
      assert(args.okdone());
      dec_ret b = (sob->*dec)(a1,a2,a3,r,wccret);
      ret << r;
      return b;
    }
  };
  rsm_handler * hp =  new h1(sob, meth, dec);
  if(dec) hp->setusedecision(true);
  reg1(proc,hp);
}

// Testing Decision function addition
template<class S, class A1, class A2, class A3, class A4, class R> void
  rsm::reg(int proc, S*sob, int (S::*meth)(const A1 a1, const A2 a2,
             const A3 a3, const A4 a4, R & r),
             dec_ret (S::*dec)(const A1 a1, const A2 a2, const A3 a3, const A4 a4, R & r, int & r2))
{
 class h1 : public rsm_handler {
  private:
    S * sob;
    int (S::*meth)(const A1 a1, const A2 a2, const A3 a3, const A4 a4, R & r);
    dec_ret (S::*dec)(const A1 a1, const A2 a2, const A3 a3, const A4 a4, R & r, int & r2);
  public:
  h1(S *xsob, int (S::*xmeth)(const A1 a1, const A2 a2, const A3 a3,
            const A4 a4, R & r),dec_ret (S::*xdec)(const A1 a1, const A2 a2, const A3 a3,
                    const A4 a4, R & r, int & r2))
    : sob(xsob), meth(xmeth), dec(xdec) { }
    int fn(unmarshall &args, marshall &ret) {
      A1 a1;
      A2 a2;
      A3 a3;
      A4 a4;
      R r;
      args >> a1;
      args >> a2;
      args >> a3;
      args >> a4;
      assert(args.okdone());
      int b = (sob->*meth)(a1,a2,a3,a4,r);
      ret << r;
      return b;
    }
    dec_ret decide(unmarshall &args, marshall &ret, int &wccret) {
      if(!dec) return YES_REP;
      A1 a1;
      A2 a2;
      A3 a3;
      A4 a4;
      R r;
      args >> a1;
      args >> a2;
      args >> a3;
      args >> a4;
      assert(args.okdone());
      dec_ret b = (sob->*dec)(a1,a2,a3,a4,r,wccret);
      ret << r;
      return b;
    }
  };
  rsm_handler * hp =  new h1(sob, meth, dec);
  if(dec) hp->setusedecision(true);
  reg1(proc,hp);
}


template<class S, class A1, class A2, class A3, class A4, class A5, class R> void
  rsm::reg(int proc, S*sob, int (S::*meth)(const A1 a1, const A2 a2,
             const A3 a3, const A4 a4,
             const A5 a5, R & r)
             ,dec_ret (S::*dec)(const A1 a1, const A2 a2, const A3 a3, const A4 a4,const A5 a5, R & r, int & r2))
{
 class h1 : public rsm_handler {
  private:
    S * sob;
    int (S::*meth)(const A1 a1, const A2 a2, const A3 a3, const A4 a4,
       const A5 a5, R & r);
    dec_ret (S::*dec)(const A1 a1, const A2 a2, const A3 a3, const A4 a4,const A5 a5, R & r, int & r2);
  public:
  h1(S *xsob, int (S::*xmeth)(const A1 a1, const A2 a2, const A3 a3,
            const A4 a4, const A5 a5, R & r),dec_ret (S::*xdec)(const A1 a1, const A2 a2, const A3 a3,
                    const A4 a4,const A5 a5, R & r, int & r2))
    : sob(xsob), meth(xmeth), dec(xdec) { }
    int fn(unmarshall &args, marshall &ret) {
      A1 a1;
      A2 a2;
      A3 a3;
      A4 a4;
      A5 a5;
      R r;
      args >> a1;
      args >> a2;
      args >> a3;
      args >> a4;
      args >> a5;
      assert(args.okdone());
      int b = (sob->*meth)(a1,a2,a3,a4,a5,r);
      ret << r;
      return b;
    }
    dec_ret decide(unmarshall &args, marshall &ret, int &wccret) {
      if(!dec) return YES_REP;
      A1 a1;
      A2 a2;
      A3 a3;
      A4 a4;
      A5 a5;
      R r;
      args >> a1;
      args >> a2;
      args >> a3;
      args >> a4;
      args >> a5;
      assert(args.okdone());
      dec_ret b = (sob->*dec)(a1,a2,a3,a4,a5,r,wccret);
      ret << r;
      return b;
    }
  };
  rsm_handler * hp =  new h1(sob, meth, dec);
  if(dec) hp->setusedecision(true);
  reg1(proc,hp);
}

#endif /* rsm_h */
