/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang,
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


#include <assert.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include "rsm_log.h"
#include "wfs_misc.h"

enum {
  LOGSZ_TO_CHECKPOINT = 10*1024 //in bytes
};

rsm_log::rsm_log(std::string _me, std::string log_dir) :
  rfd(-1),last_cp_vs(0,0)
{
	lname = log_dir + "/rsm-" + _me + ".app.log";
	cname = log_dir + "/rsm-" + _me + ".cp.log";
	wfd = open(lname.c_str(), O_APPEND|O_CREAT|O_WRONLY,0777);
	if (wfd < 0) {
		perror("rsm_log open append log error");
		assert(0);
	}else{

		wfs_log(WFS_DBG_3,"rsm_log append to fd %d\n", wfd);

	}
	total = lseek(wfd, 0, SEEK_END);
}

rsm_log::~rsm_log()
{
	if (rfd >=0)
		close(rfd);
	if (wfd >=0)
		close(wfd);
}

void
rsm_log::write_record(int fd, const rsm_record &r, bool sync)
{
	int n;
	n = write(fd, &r.vs, sizeof(r.vs));
	if (n < 0) {
		perror("rsm_log write error:");
		wfs_log(WFS_DBG_3,"rsm_log write error: fd is %d\n", fd);
	}
	assert(n == sizeof(r.vs));
	n = write(fd, &r.procno, sizeof(r.procno));
	assert(n==sizeof(r.procno));
	int rsz = r.req.length();
	n = write(fd, &rsz, sizeof(rsz));
	assert(n==sizeof(rsz));
	n = write(fd, r.req.c_str(), rsz);
	assert(n==rsz);
	total += sizeof(r.vs)+sizeof(r.procno)+sizeof(rsz)+rsz;

	wfs_log(WFS_DBG_3,"rsm_log::write_record %d,%d procno %x reqsz %u total %lu sync %d\n", r.vs.vid,r.vs.seqno,r.procno,(unsigned)r.req.length(),total,sync);

	if(sync) {
#if defined(__FreeBSD__) || defined(__APPLE__)
	  assert(fsync(fd)==0);
#else
	  assert(fdatasync(fd)==0);
#endif
	}

	wfs_log(WFS_DBG_3,"rsm_log::write_record done %d,%d procno %x reqsz %u total %lu\n", r.vs.vid,r.vs.seqno,r.procno,(unsigned)r.req.length(),total);

}

bool
rsm_log::read_record(int fd, rsm_record &r)
{
	int rsz;
	int n;
	r.vs.vid = r.vs.seqno = r.procno = 0;

	if( fd < 0 ) {
	  return false;
	}

	n = read(fd, &r.vs, sizeof(r.vs));
	if (n == 0)
		return false;
	if (n < 0) {
		perror("rsm_log read failed");
	}
	assert(n==sizeof(r.vs));
	n = read(fd, &r.procno, sizeof(r.procno));
	assert(n==sizeof(r.procno));

	n = read(fd, &rsz, sizeof(rsz));
	assert(n==sizeof(rsz));

	char *b = (char *)malloc(rsz);
	n = read(fd, b, rsz);
	assert(n==rsz);

	wfs_log(WFS_DBG_3,"rsm_log::read_record %d,%d procno %x reqsz %d\n", r.vs.vid, r.vs.seqno, r.procno, rsz);

	r.req = std::string(b,rsz);
	free(b);
	return true;
}

void
rsm_log::reset_read()
{
	if (wfd >= 0) {
		close(wfd);
		wfd = -1;
	}
	if( rfd >= 0 ) {
	  close(rfd);
	}
	rfd = open(lname.c_str(), O_RDONLY);
	lseek(rfd, 0, SEEK_SET);

	wfs_log(WFS_DBG_3,"rsm_log::reset_read %s fd %d\n", lname.c_str(), rfd);
}

void
rsm_log::reset_write()
{
	wfd = open(lname.c_str(), O_CREAT|O_APPEND|O_WRONLY,0777);
	total += lseek(wfd, 0, SEEK_END);

	wfs_log(WFS_DBG_3,"rsm_log::reset_write fd %d sz %lu\n", wfd,total);

}

void
rsm_log::save_request(const rsm_record &r, bool sync)
{
	if (wfd < 0)
		reset_write();
	write_record(wfd, r, sync);
}

bool
rsm_log::load_request(const viewstamp &vs, rsm_record &r)
{
	if (rfd < 0) {
		reset_read();
	}
	do {
		if (!read_record(rfd, r))
			return false;;
	}while (vs > r.vs);
	return true;
}

bool
rsm_log::do_checkpoint()
{

  wfs_log(WFS_DBG_3,"rsm_log::do_checkpoint? bytes written since last cp: %lu\n", total);
  return (total > LOGSZ_TO_CHECKPOINT);
}

void
rsm_log::save_checkpoint(const rsm_record &r)
{
	std::string ctmp = cname + ".tmp";
	int cp_fd = open(ctmp.c_str(),O_CREAT|O_TRUNC|O_WRONLY,0777);
	write_record(cp_fd, r);
	close(cp_fd);
	rename(ctmp.c_str(),cname.c_str());
  last_cp_vs = r.vs;

  //truncate the log entirely
  close(wfd);
	wfd = open(lname.c_str(), O_TRUNC|O_WRONLY,0777);
  total = 0;
  assert(wfd>=0);


  wfs_log(WFS_DBG_3,"rsm_log::save_checkpoint saved vs %d,%d size %u\n",
	   r.vs.vid, r.vs.seqno, (unsigned)r.req.length());
}

bool
rsm_log::load_checkpoint(rsm_record &r)
{
	int cp_fd = open(cname.c_str(),O_RDONLY);
  if (cp_fd < 0)
    return false;
	bool ret = read_record(cp_fd, r);
	close(cp_fd);
  if (r.vs > last_cp_vs)
    last_cp_vs = r.vs;


  wfs_log(WFS_DBG_3,"rsm_log::load_checkpoint load vs %d,%d size %u\n",
	   last_cp_vs.vid, last_cp_vs.seqno, (unsigned)r.req.length());

  return ret;
}
