/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang,
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


/*
 * slice_manager.cc
 */
#include <method_thread.h>
#include "slice_manager.h"

// Not used anymore
bool
compare_by_fid_count(slice_manager::slice_ent * pFirst,slice_manager::slice_ent * pSecond)
{
  return *pFirst < *pSecond;
}

slice_manager::slice_manager(bool from_log, std::string xdst, host me,
			     wfs_client_manager *clients,
			     int sync_period, int lease_period)
  : _me(me), _master(xdst), _try_root_slice(true), _sync_period(sync_period), _shutting_down(false), _lease_period(lease_period)
{
  pthread_mutex_init(&_slices_mutex, NULL);
  pthread_mutex_init(&_sites_mutex, NULL);
  _ccl =  new wfs_rsm_cclient(_master,_me,clients);
  method_thread( this, true, &slice_manager::loop );
}

slice_manager::~slice_manager()
{
  _shutting_down = true;
  pthread_mutex_destroy(&_slices_mutex);
  pthread_mutex_destroy(&_sites_mutex);
  delete _ccl;
}

void slice_manager::loop() {

  sleep(1);

  while(!_shutting_down) {
    sync_state();
    sleep(_sync_period);
  }
}

void slice_manager::slice_update(slice_ent * pEnt, bool released){

  wfs_log(WFS_DBG_3,"Slice ID:%016"PRIx64" Needs Update. New state is:%s\n",
        pEnt->sa.slice.val(),pEnt->sa.to_canonical_string().c_str());

  if(!pEnt->locked){
    // Don't add to data structure.

    wfs_log(WFS_DBG_3,"Slice ID:%016"PRIx64" Needs Update. Not locked so returning.\n",
	     pEnt->sa.slice.val());

    return;
  }
  // insert to known policies
  policymap::iterator iter3;
  if((iter3 = _sat_policies.find(make_key_from_pol(pEnt->sa.policy))) != _sat_policies.end()){
    iter3->second.insert(pEnt);
  } else {
    std::set<slice_ent *> newset;
    newset.insert(pEnt);
    std::string key = make_key_from_pol(pEnt->sa.policy);
    _sat_policies[key] = newset;
  }
  // If we are primaries for this slice, see if it complies with the policy.
  // If not, it will be marked and fixed during the next call to check_all_slices.
  if(pEnt->amiprimary(_me)){
    does_entry_comply(pEnt);
  }
  // Add see if there are new hosts to add.
  std::list<host>::iterator lst_it;
  sitemap::iterator site_it;
  assert(pthread_mutex_lock(&_sites_mutex) == 0);
  for(lst_it = pEnt->sa.replicas.begin();lst_it != pEnt->sa.replicas.end();++lst_it ){
    host h = *lst_it;
    if(h == _me) // This host should never be in the sitemap.
      continue;
    if((site_it = _sites.find(h.site_id)) == _sites.end()){
      // Add with a new site.

      wfs_log(WFS_DBG_3,"slice_manager::slice_update:: Got new site:%s from cserver\n",h.site_id.c_str());

      std::set<host> newset;
      newset.insert(h);
      //_sites.insert(std::pair<std::string,std::set<host> >(h.site_id,newset));
      _sites[h.site_id] = newset;

    } else {
      // See if should add to existing site.
      if(site_it->second.find(h) == site_it->second.end()){

        wfs_log(WFS_DBG_3,"slice_manager::slice_update:: Got new host:%s in site:%s from cserver\n"
              ,(static_cast<std::string>(h)).c_str(),h.site_id.c_str());

        site_it->second.insert(h);
      }
    }
  }
  assert(pthread_mutex_unlock(&_sites_mutex) == 0);
}

bool slice_manager::get_slices_updates()
{
  slicemap::iterator iter;
  slice_ent * pEnt;
  std::vector<slice_lock> locks;
  viewstamp ret;

  assert(pthread_mutex_lock(&_slices_mutex) == 0);
  int r  = _ccl->fetch_slices(_update_stamp,locks,ret);
  if(r == wcc_protocol::OK){

    wfs_log(WFS_DBG_3,"slice_manager: Old Viewstamp:(%d,%d) New Viewstamp:(%d,%d) Got %u Entries\n"
          ,_update_stamp.vid,_update_stamp.seqno,ret.vid,ret.seqno,
	     (unsigned) locks.size());

    _update_stamp = ret;
  } else {

    wfs_log(WFS_DBG_1,"slice_manager: Error during synchronization process with configuration server. Error Code:%d\n",r);

    assert(pthread_mutex_unlock(&_slices_mutex) == 0);
    return false;
  }
  // Convert those locks into slice assignments.
  for( uint i = 0; i < locks.size(); i++ ) {
	wfs_slice_id sid = locks[i].name;
    bool stateChanged = false,released = false;
    if((iter = _slices.find(sid)) != _slices.end()){
      // Known slices.
      pEnt = iter->second;
      // Need to update locks that expired (no state change). Free locks that were locked
      // will be observed since their state changes. Currently the server sends the old state for expired locks.
      if((released = (pEnt->locked && (locks[i].avail == slice_lock::FREE)))
          || (pEnt->lock_state != locks[i].state)) {
        stateChanged = true;
        pEnt->lock_state = locks[i].state;
        pEnt->sa.from_canonical_string(pEnt->lock_state);
      }
    } else {
      // New slices
      stateChanged = true;
      pEnt = new slice_ent(locks[i].state);
      pEnt->lock_state = locks[i].state;
      _slices[sid] = pEnt;
    }
    // Always update whether or not locked.
    pEnt->locked = (locks[i].avail == slice_lock::OWNED);
    // See if we need to sync other data structures.
    if(stateChanged){
      slice_update(pEnt,released);
    }
  }
  _last_sync = time(NULL);
  assert(pthread_mutex_unlock(&_slices_mutex) == 0);
  return true;
}

bool slice_manager::sync_state()
{
  return get_slices_updates();
}

bool slice_manager::is_primary_for_file(const wfs_file_id fid) const
{
  bool ret;
  slice_ent * pEnt;
  slicemap::const_iterator iter;
  assert(pthread_mutex_lock(&_slices_mutex) == 0);
  if((iter = _slices.find( wfs_slice::get_slice_id(fid))) == _slices.end()){
    ret = false;
  } else {
	pEnt = iter->second;
    ret = pEnt->amiprimary(_me) && pEnt->islocked(_ccl,_me);
  }
  assert(pthread_mutex_unlock(&_slices_mutex) == 0);
  return ret;
}

slice_manager::ret_code
slice_manager::get_slice_primary(const wfs_slice_id sid, host & prim, bool onlyIfLocked)
{
  ret_code ret;
  bool found_slice = false;
  slice_ent * pEnt;
  slicemap::const_iterator iter;
  assert(pthread_mutex_lock(&_slices_mutex) == 0);
  if((iter = _slices.find(sid)) == _slices.end()){
    // Try and sync with the server looking for that slice
    assert(pthread_mutex_unlock(&_slices_mutex) == 0);
    //TODO: Replace this with an rsm client call get_single_lock;
    sync_state();
    assert(pthread_mutex_lock(&_slices_mutex) == 0);
    if((iter = _slices.find(sid)) == _slices.end()){
      ret = SM_NOENT;
    } else{
      found_slice = true;
    }
  } else {
    found_slice = true;
  }
  if(found_slice) {
    pEnt = iter->second;
    if(onlyIfLocked) {
      if(pEnt->islocked(_ccl,_me)){
        prim = pEnt->primary();
        ret = SM_SUCCESS;
      } else {
    	ret = SM_NOLOCKED;
      }
    } else {
      prim = pEnt->primary();
      ret = SM_SUCCESS;
    }
  }
  assert(pthread_mutex_unlock(&_slices_mutex) == 0);
  return ret;
}

slice_manager::ret_code
slice_manager::get_slice_replicaset(const wfs_slice_id sid, std::vector<host>& repv, int & rep_num)
{
  ret_code ret;
  slice_ent * pEnt;
  slicemap::const_iterator iter;
  bool found_slice = false;
  assert(pthread_mutex_lock(&_slices_mutex) == 0);
  if((iter = _slices.find(sid)) == _slices.end()){
    // Try and sync with the server looking for that slice
    assert(pthread_mutex_unlock(&_slices_mutex) == 0);
    //TODO: Replace this with an rsm client call get_single_lock;
    sync_state();
    assert(pthread_mutex_lock(&_slices_mutex) == 0);
    if((iter = _slices.find(sid)) == _slices.end()){
      ret = SM_NOENT;
    } else{
      found_slice = true;
    }
  } else {
    found_slice = true;
  }

  if(found_slice){
    pEnt = iter->second;
    //no need for clear(). assign does that.
    repv.assign(pEnt->sa.replicas.begin(),pEnt->sa.replicas.end());
    rep_num = pEnt->sa.policy.rep_num;
    ret = SM_SUCCESS;
  }
  assert(pthread_mutex_unlock(&_slices_mutex) == 0);
  return ret;
}
// Got to have _slices_mutex locked to call
slice_manager::ret_code
slice_manager::find_unused_slice(wfs_slice_id & sid)
{
  // Make sure we have the latest slice data from the cserver.
  // Even if we fail try and use last known state. Inconsistencies will be caught when we will try and
  // lock the slice.
  wfs_slice_id newsid;
  assert(pthread_mutex_unlock(&_slices_mutex) == 0);
  if(!sync_state()){
    wfs_log(WFS_DBG_1,"slice_manager::find_unused_slice. Warning: Cannot sync with cserver. State may be stale.\n");
  }
  assert(pthread_mutex_lock(&_slices_mutex) == 0);

  // TODO: handle switch to permutation here
  if(true || _usedslices_count < wfs_slice::MAX_SLICE){
    do {
      newsid = random() % wfs_slice::MAX_SLICE;
      newsid <<= (sizeof(wfs_slice_id)*8) - wfs_slice::SLICE_BIT_COUNT;
    } while(_slices.find(newsid) != _slices.end());
  } else {
    // TODO: Use permutation.
  }

  wfs_log(WFS_DBG_3,"slice_manager::find_unused_slice. Selected new slice ID:%016"PRIx64".\n",newsid);

  _usedslices_count++;
  sid = newsid;

  return SM_SUCCESS;
}

// NOTE: not used anymore in create_new_fid. the gain in those rare scenarios
// is not worth the code complexity. Y.S.
slice_manager::ret_code
slice_manager::duplicate_slice_assignment(const slice_ent * copyfrom,slice_ent ** ppEnt)
{
  ret_code ret = SM_SUCCESS;
  int rsmret = wcc_protocol::RETRY;
  wfs_slice_id newsid;
  slice_ent * newEnt = NULL;
  // Try with a new slice if this one is locked.
  do {
    if(newEnt) delete newEnt;
    if((ret = find_unused_slice(newsid)) != SM_SUCCESS){
      return ret;
    }
    wfs_slice newslice(newsid);
    newEnt = new slice_ent(newslice,copyfrom->sa.policy,copyfrom->sa.replicas);

    rsmret = _ccl->acquire_slice(&newEnt->sa,_lease_period,true);
    if((rsmret != wcc_protocol::OK) && (rsmret != wcc_protocol::RETRY)){
      delete newEnt;

      wfs_log(WFS_DBG_1,"slice_manager::duplicate_slice_assignment. Cannot lock slice:%016"PRIx64". WCC_ERROR:%d\n"
               ,newslice.val(),rsmret);

      return SM_RSM_ERR;
    }

  } while (rsmret == wcc_protocol::RETRY);
  // Update state
  newEnt->lock_state = newEnt->sa.to_canonical_string();
  *ppEnt = newEnt;
  return ret;
}

// Think of i as the number of replicas. that already have been added.
slice_manager::ret_code
slice_manager::get_random_replica(int i, int rep_sites, std::set<std::string>& sites_used,const std::vector<std::string>& sites_vec,
    std::set<host> nodes_used , std::map<std::string,std::pair<int,int> >& nodes_per_site, host & h, std::string& s)
{
  sitemap::const_iterator it;
  std::set<host>::const_iterator it3;
  int known_nodes = 0;
  bool insist_unused = true, found_site = false;
  s = "";

  for(it = _sites.begin(); it != _sites.end(); ++it){
    known_nodes+= it->second.size();
  }
  if (i > known_nodes){

    wfs_log(WFS_DBG_2,"slice_manager::create_new_slice_entry. Requested %d replicas but only %d are known\n"
          ,i+1,known_nodes);

    return SM_GENERAL_ERR;
  }
  if (i >= rep_sites) {
    // Maybe all used sites are full. Check for this.
    int m = i%rep_sites;
    for(unsigned l = 0;l < sites_vec.size();++l,++m){
      if(nodes_per_site[sites_vec[m % sites_vec.size()]].first <
          nodes_per_site[sites_vec[m % sites_vec.size()]].second){
        s = sites_vec[m % sites_vec.size()];

        wfs_log(WFS_DBG_3,"slice_manager::create_new_slice_entry. Adding host from site:%s\n"
              ,s.c_str());

        found_site = true;
        break;
      }
    }
  }
  if(!found_site){
    int k=0;
    do {
      // Choose a random site.
      bool skip = true;
      unsigned j = 0;
      unsigned offset = random() % _sites.size();
      for(it = _sites.begin(); j < _sites.size();++j){
        if(skip){
          if(j < offset) {
            ++it;
            continue;
          } else {
            // We are at the random offset. Start a full cycle.
            skip = false;
            j = 0;
          }
        }
        if((nodes_per_site[it->first].first < nodes_per_site[it->first].second) // has unused nodes.
           && (!insist_unused  || sites_used.find(it->first) == sites_used.end())){   // unused site
          s = it->first;

          wfs_log(WFS_DBG_3,"slice_manager::create_new_slice_entry. Adding a host from site:%s\n"
                ,s.c_str());

          found_site = true;
          break;
        }
        // Continue iterating
        ++it;
        if(it == _sites.end())
          it = _sites.begin();
      }
      if(j == _sites.size()){
        // Could not find an unused site with nodes to give.
        // But there must be a used site with unused nodes since i <= known_nodes

	    wfs_log(WFS_DBG_1,"Can't find unused %d\n",i);
        insist_unused = false;
      }
      k++;
    } while (!found_site && k < 2);
  }

  // couldn't find anything
  if( s == "" ) {
    return SM_GENERAL_ERR;
  }

  // Should be able to find a random node
  std::set<host> * pHosts = &(_sites[s]);
  unsigned offset = random() % pHosts->size();

  bool skip = true;
  unsigned j = 0;
  for(it3 = pHosts->begin(); j < pHosts->size();++j){
    if(skip) {
      if(j < offset){
        ++it3;
        continue;
      } else {
        // We are at the random offset. Start a full cycle.
        skip = false;
        j = 0;
      }
    }

    if(nodes_used.find(*it3) == nodes_used.end()){
      h = *it3;
      wfs_log(WFS_DBG_3,"slice_manager::create_new_slice_entry. Adding a host:%s\n"
            ,h.str().c_str());
      break;
    }
    ++it3;
    if(it3 == pHosts->end())
      it3 = pHosts->begin();
  }
  assert(j < pHosts->size() && (s == h.site_id));
  return SM_SUCCESS;
}


slice_manager::ret_code
slice_manager::create_new_slice_entry(const rep_policy& pol, slice_ent ** ppEnt,
				      wfs_slice_id forced_id,
				      std::vector<host> forced_replicas)
{
  int rsmret;
  ret_code ret = SM_SUCCESS,ret2 = SM_SUCCESS;
  slice_ent * newEnt = NULL;
  std::list<host> replicas;
  wfs_slice_id newsid;
  int remaining_hosts = pol.rep_num, remaining_sites = pol.site_num;
  int rep_level = pol.rep_num, rep_sites = pol.site_num;
  std::set<host> nodes_used;
  std::set<std::string> sites_used;
  std::vector<std::string> sites_vec; // We use this to get a consistent iteration order.
  // Counts how many nodes are available from each site and how many have been assigned from each site.
  // nodes_per_site[site_id] = <used,availabe>
  std::map<std::string,std::pair<int,int> > nodes_per_site;
  host h;
  std::string s;

  if(pol.rep_num < 0 || pol.rep_num < pol.site_num)
    return SM_GENERAL_ERR;

  assert(pthread_mutex_lock(&_sites_mutex) == 0);
  for(sitemap::const_iterator it = _sites.begin(); it != _sites.end();++it){
    nodes_per_site[it->first] = std::pair<int,int>(0,it->second.size());
  }
  // This host is never in _sites. Need to count it in capacity too.
  if(nodes_per_site.find(_me.site_id) != nodes_per_site.end())
    nodes_per_site[_me.site_id].second++;
  assert(nodes_per_site.size() == _sites.size());

  nodes_used.insert(_me);
  replicas.push_back(_me);
  sites_used.insert(_me.site_id);
  sites_vec.push_back(_me.site_id);
  remaining_hosts--; remaining_sites--;
  (nodes_per_site[_me.site_id].first)++;

  for( int i = 0; i < forced_replicas.size(); i++ ) {
    if( forced_replicas[i] != _me ) {

      wfs_log(WFS_DBG_2,"Using forced replica %s\n", forced_replicas[i].str().c_str() );

      replicas.push_back(forced_replicas[i]);
      remaining_hosts--;
      (nodes_per_site[forced_replicas[i].site_id].first)++;
      if((sites_used.insert(forced_replicas[i].site_id)).second){
	// The site was inserted
	sites_vec.push_back(forced_replicas[i].site_id);
	remaining_sites--;
      }
    }
  }

  wfs_log(WFS_DBG_3,"slice_manager::create_new_slice_entry. Adding primary host:%s.Primary site:%s\n"
        ,_me.str().c_str(),_me.site_id.c_str());

  for( int i = replicas.size(); i < rep_level; i++ ) {
    ret = get_random_replica(i,rep_sites,sites_used,sites_vec,nodes_used,nodes_per_site,h,s);
    if(ret != SM_SUCCESS){
      break;
    }
    nodes_used.insert(h); //  an unused node is always returned on success
    replicas.push_back(h);
    (nodes_per_site[h.site_id].first)++;
    remaining_hosts--;
    if((sites_used.insert(s)).second){
      // The site was inserted
      sites_vec.push_back(s);
      remaining_sites--;
    }
  }
  assert(pthread_mutex_unlock(&_sites_mutex) == 0);

  // Thats the best we can do. See if we got enough hosts
  if(remaining_hosts > 0 || remaining_sites > 0){

    wfs_log(WFS_DBG_1,"slice_manager::create_new_slice_entry. Can't Fully satisfy the policy. Replication number:%d. Known hosts:%d "
          "Site Replication:%d Known Site:%d \n"
                ,pol.rep_num,pol.rep_num - remaining_hosts,pol.site_num,pol.site_num -remaining_sites );

    ret = SM_SA_INC;
  }
  do {
    if(newEnt) delete newEnt;

    if(forced_id != 1) {
      newsid = forced_id;
    } else if(_try_root_slice){
      newsid = wfs_slice::ROOT_SLICE;
      _try_root_slice = false;
    } else {
      // Choose an unused slice.
      ret2 = find_unused_slice(newsid);
      if(ret2 != SM_SUCCESS){
         return ret2;
      }
    }

    wfs_slice newslice(newsid);
    newEnt = new slice_ent(newslice,pol,replicas);

    rsmret = _ccl->acquire_slice(&newEnt->sa,_lease_period,true);
    if((rsmret != wcc_protocol::OK) && (rsmret != wcc_protocol::RETRY)){
      delete newEnt;

      wfs_log(WFS_DBG_1,"slice_manager::create_new_slice_entry. Cannot lock slice:%016"PRIx64". WCC_ERROR:%d\n"
                ,newslice.val(),rsmret);
      return SM_RSM_ERR;
    }
    if(forced_id != 1 && rsmret == wcc_protocol::RETRY){
      return SM_CANT_SATISFY;
    }
  } while (rsmret == wcc_protocol::RETRY);
  // Update state
  newEnt->lock_state = newEnt->sa.to_canonical_string();
  newEnt->locked = true;
  *ppEnt = newEnt;
  return ret;
}
#if 0
slice_manager::ret_code
slice_manager::create_new_slice_entry(const rep_policy& pol, slice_ent ** ppEnt,
                      wfs_slice_id forced_id)
{
  int rsmret;
  ret_code ret = SM_SUCCESS,ret2 = SM_SUCCESS;
  slice_ent * newEnt = NULL;
  std::list<host> replicas;
  std::set<host> accum_set;
  sitemap::iterator site_iter;
  wfs_slice_id newsid;
  int remaining_hosts = pol.rep_num, remaining_sites = pol.site_num;
  //int replicas_on_prim_site = (pol.rep_num - pol.site_num) + 1;

  assert(pol.rep_num > 0);
  assert(pol.rep_num >= pol.site_num);

  /*
   * Satisfying the policy with an assignment. I.E create the replica list.
   * This is a best effort operation. If the assignment doesn't satisfy the policy completely
   * we return an appropriate error code but still try to assign the new replica set.
   * The assignment will try and be corrected later.
   *
   * Just looking at this code gives me a headache but I can't think of a way to simplify it Y.S.
   */
  // I am the primary
  replicas.push_front(_me);
  accum_set.insert(_me);
  remaining_hosts--;
  remaining_sites--;

  wfs_log(WFS_DBG_3,"slice_manager::create_new_slice_entry. Adding primary host:%s.Primary site:%s As replica number:%d\n"
            ,(static_cast<std::string>(_me)).c_str(),_me.site_id.c_str(),pol.rep_num -remaining_hosts );

  assert(pthread_mutex_lock(&_sites_mutex) == 0);
  // Invariant remaining_hosts >= remaining_sites helps us here.
  for(site_iter = _sites.begin();(remaining_sites > 0) && (site_iter != _sites.end());++site_iter){
    if((site_iter->first != _me.site_id) && (site_iter->second.size() > 0)){
      host h = *(site_iter->second.begin());
      replicas.push_back(h);
      accum_set.insert(h);
      remaining_hosts--;
      remaining_sites--;

      wfs_log(WFS_DBG_3,"slice_manager::create_new_slice_entry. Adding host:%s from site:%s As replica number:%d\n"
                ,static_cast<std::string>(h).c_str(),h.site_id.c_str(),pol.rep_num - remaining_hosts);
    }
  }
  if(remaining_sites > 0){
    // This means that we don't have enough sites to satisfy the site placement policy.
    wfs_log(WFS_DBG_3,"slice_manager::create_new_slice_entry. Can't Fully satisfy the policy. Requested sites:%d. Known sites:%d\n"
              ,pol.site_num,pol.site_num - remaining_sites);
    ret = SM_SA_INC;
  }
  // Try to push the rest of the hosts from the primary site.
  if((_me.site_id != "") &&
     ((site_iter = _sites.find(_me.site_id)) != _sites.end()) &&
     (site_iter->second.size() > 0)){
    for(std::set<host>::iterator it = site_iter->second.begin();
        (remaining_hosts > 0) && ( it != site_iter->second.end());++it){
      if(accum_set.find(*it) != accum_set.end()){
          // Already have it
          continue;
      }
      host h = *it;
      replicas.push_back(h);
      accum_set.insert(h);
      remaining_hosts--;
      remaining_sites--;

      wfs_log(WFS_DBG_3,"slice_manager::create_new_slice_entry. Adding host:%s from site:%s As replica number:%d\n"
                ,static_cast<std::string>(h).c_str(),h.site_id.c_str(),pol.rep_num - remaining_hosts);
    }
  }
  // Try to push the rest of the host from where ever.
  for(site_iter = _sites.begin();(remaining_hosts > 0) && (site_iter != _sites.end());++site_iter){
    if(site_iter->first == _me.site_id)
      continue;
    for(std::set<host>::iterator it = site_iter->second.begin();
            (remaining_hosts > 0) && ( it != site_iter->second.end());++it){
      if(accum_set.find(*it) != accum_set.end()){
        // Already have it
        continue;
      }
      host h = *it;
      replicas.push_back(h);
      accum_set.insert(h);
      remaining_hosts--;
      remaining_sites--;

      wfs_log(WFS_DBG_3,"slice_manager::create_new_slice_entry. Adding host:%s from site:%s As replica number:%d\n"
            ,static_cast<std::string>(h).c_str(),h.site_id.c_str(),pol.rep_num - remaining_hosts);

    }
  }
  assert(pthread_mutex_unlock(&_sites_mutex) == 0);
  accum_set.clear();
  // Thats the best we can do. See if we got enough hosts
  if(remaining_hosts > 0){

    wfs_log(WFS_DBG_3,"slice_manager::create_new_slice_entry. Can't Fully satisfy the policy. Requested replication:%d. Known hosts:%d\n"
                ,pol.rep_num,pol.rep_num - remaining_hosts);

    ret = SM_SA_INC;
  }

  do {
    if(newEnt) delete newEnt;

    if(forced_id != 1) {
      newsid = forced_id;
    } else if(_try_root_slice){
      newsid = wfs_slice::ROOT_SLICE;
      _try_root_slice = false;
    } else {
      // Choose an unused slice.
      ret2 = find_unused_slice(newsid);
      if(ret2 != SM_SUCCESS){
         return ret2;
      }
    }

    wfs_slice newslice(newsid);
    newEnt = new slice_ent(newslice,pol,replicas);

    rsmret = _ccl->acquire_slice(&newEnt->sa,_lease_period,true);
    if((rsmret != wcc_protocol::OK) && (rsmret != wcc_protocol::RETRY)){
      delete newEnt;

      wfs_log(WFS_DBG_3,"slice_manager::create_new_slice_entry. Cannot lock slice:%016"PRIx64". WCC_ERROR:%d\n"
                ,newslice.val(),rsmret);

      return SM_RSM_ERR;
    }
    if(forced_id != 1 && rsmret == wcc_protocol::RETRY){
      return SM_CANT_SATISFY;
    }
  } while (rsmret == wcc_protocol::RETRY);
  // Update state
  newEnt->lock_state = newEnt->sa.to_canonical_string();
  newEnt->locked = true;
  *ppEnt = newEnt;
  return ret;
}
#endif

template<class T>
static bool get_random_entry_from_set(std::set<T>& tset,T& t)
{
  if(tset.size() == 0)
    return false;

  typename std::set<T>::iterator it;
  int offset = random() % tset.size();
  int i = 0;
  for (it = tset.begin();it != tset.end(); ++it,++i) {
    if(i == offset){
      t = *it;
      return true;
    }
  }
  assert (false);
}

/*  LOGIC FOR get_new_fid
slice_set s = set of slices satisfying the requested policy
if( s is empty ) {
   if( local site == policy's primary site ) {
      create_new_slice where primary = local host
      choose fid from the new slice
      return SM_SUCCESS
   } else {
      try_host = random node at the policy's primary site
      return SM_TRY_HOST
   }
} else {
   if( s contains a slice S where the local host is primary ) {
      choose fid from S
      return SM_SUCCESS
   } else if( force_create ) {
      create_new_slice where primary = local host
      choose fid from the new slice
      return SM_SUCCESS
   } else {
      try_host = primary of a random slice in s
      return SM_TRY_HOST
   }
}
*/
// TODO: Once we settle on the correct logic - this method needs some major refactoring. Y.S.
// Get rid of all the ugly booleans
slice_manager::ret_code
  slice_manager::get_new_fid(const rep_policy& pol,wfs_file_id& fid, host& tryhost,bool force_local_primary, wfs_slice_id forced_id, std::vector<host> forced_replicas)
{

  policymap::iterator iter;
  std::set<slice_ent *>::iterator it;
  sitemap::iterator iter2;
  slice_ent * pEnt = NULL, * pNent = NULL;
  int r;
  ret_code ret = SM_SUCCESS;

  // Most commonly we will be asked for fids from seen-before policies. so first check to see if we already
  // have a slice satisfying this policy.
  bool have_slice = false;
  assert(pthread_mutex_lock(&_slices_mutex) == 0);
  if((iter = _sat_policies.find(make_key_from_pol(pol))) != _sat_policies.end()){
    // If there is a known existing policy, there are 3 cases.
    for (it = iter->second.begin();it != iter->second.end(); ++it){
      if((*it)->amiprimary(_me)){
        // Case A: found a slice for this policy where the local host is primary.
	if(forced_id != 1 && (*it)->sa.slice.val() != forced_id) {
	  continue;
        } else if(!(*it)->isfull()){
          pEnt = *it;
          have_slice = true;

          wfs_log(WFS_DBG_3,"slice_manager::get_new_fid. Existing slice:%016"PRIx64" satisfies the requested policy\n"
                ,pEnt->sa.slice.val());

          break;
        } else {
          // If at least one good slice was found but it was full - need to force a creation of a new slice.
          force_local_primary = true;
        }
      }
    }
    // Can't serve from existing slices.
    if(!have_slice){
      if(force_local_primary){
        // Case B: This will force a new slice to be created and this host will be pushed as the primary.
        have_slice = false;

        wfs_log(WFS_DBG_3,"slice_manager::get_new_fid. FORCE CREATE: Creating a slice with this host as primary. Requested site:%s\n"
              ,pol.prim_site.c_str());

      } else {
        // Case C: choose a primary of a random slice in the existing policies.
        slice_ent * randEnt;
        if(get_random_entry_from_set<slice_ent *>(iter->second,randEnt)){
          tryhost = randEnt->primary();
          ret = SM_TRYHOST;

          wfs_log(WFS_DBG_3,"slice_manager::get_new_fid. Suggesting host:%s on site:%s\n",
                tryhost.str().c_str(),tryhost.site_id.c_str());

        } else {
          ret = SM_GENERAL_ERR;
        }
        assert(pthread_mutex_unlock(&_slices_mutex) == 0);
        return ret;
      }
    }
  }
  assert(pthread_mutex_unlock(&_slices_mutex) == 0);
  // HAVE A POLICY - RETURN PATH
  if(have_slice){
    // Make sure this slice is locked.
    if(!pEnt->islocked(_ccl,_me)){

      wfs_log(WFS_DBG_3,"slice_manager::get_new_fid. Requesting a lock on slice:%016"PRIx64"\n"
            ,pEnt->sa.slice.val());

      r = _ccl->acquire_slice(&pEnt->sa,_lease_period,true);
      if(r != wcc_protocol::OK){
        wfs_log(WFS_DBG_2,"slice_manager::get_new_fid. RSM client failed to acquire a lock for slice:%016"PRIx64""
              " Error code:%d\n"
              ,pEnt->sa.slice.val(),r);
	// just give up and make a new slice
	have_slice = false;
	goto dont_have;
      }
    }
    fid = pEnt->get_fid();
    return SM_SUCCESS;
  }

 dont_have:
  // *** NEW POLICY ****
  // Check to see if policy requires matching the primary site.
  // Empty string is taken as ANY site. Check if primary's site is different than this site.
  if(!have_slice &&
      (pol.site_pol == rep_policy::STRICT || pol.site_pol == rep_policy::LOCAL) &&
       pol.prim_site != ""){

    if(pol.prim_site != _me.site_id){

      wfs_log(WFS_DBG_2,"slice_manager::get_new_fid: Policy requires a primary on a different site.\n");

      // Look for a host in that site.
      assert(pthread_mutex_lock(&_sites_mutex) == 0);
      if( (iter2 = _sites.find(pol.prim_site)) == _sites.end()  || iter2->second.size() == 0 ){

        wfs_log(WFS_DBG_1,"slice_manager::get_new_fid: Can't satisfy policy. No known hosts for site:%s\n"
              ,pol.prim_site.c_str());

        ret = SM_CANT_SATISFY;
      } else {
        // Choose a random host from that site.
        if(get_random_entry_from_set<host>(iter2->second,tryhost)){
          ret = SM_TRYHOST;
        } else {
          ret = SM_GENERAL_ERR;
        }
      }
      assert(pthread_mutex_unlock(&_sites_mutex) == 0);
      return ret;
    }
  }

  // can't call on a non-server
  if (_me.addr == 0) {
    return SM_GENERAL_ERR;
  }

  /*** CREATE AN ASSIGNMENT FOR THIS POLICY *****/
  assert(pthread_mutex_lock(&_slices_mutex) == 0);
  ret = create_new_slice_entry(pol,&pNent,forced_id,forced_replicas);
  if(ret != SM_SUCCESS && ret != SM_SA_INC ){
    assert(pthread_mutex_unlock(&_slices_mutex) == 0);
    return ret;
  } else if(ret == SM_SA_INC){
    pNent->compliance = slice_ent::REPAIR;
  } else {
    pNent->compliance = slice_ent::OK;
  }
  pNent->locked = true;
  // Put the slice in the map
  _slices[pNent->sa.slice.val()] = pNent;

  // Update the policy map
  policymap::iterator iter3;
  if((iter3 = _sat_policies.find(make_key_from_pol(pNent->sa.policy))) != _sat_policies.end()){
    iter3->second.insert(pNent);
  } else {
    std::set<slice_ent *> newset;
    newset.insert(pNent);
    std::string key = make_key_from_pol(pNent->sa.policy);
    _sat_policies[key] = newset;
  }
  fid = pNent->get_fid();
  assert(pthread_mutex_unlock(&_slices_mutex) == 0);

  return ret;

}

slice_manager::ret_code
slice_manager::get_active_set(std::set<host> & accum_set)
{
  slicemap::const_iterator iter;
  std::list<host>::const_iterator  it;
  const slice_ent * pEnt;
  assert(pthread_mutex_lock(&_slices_mutex) == 0);
  for(iter=_slices.begin(); iter != _slices.end(); ++iter){
    pEnt = iter->second;
    if(!pEnt->islocked(_ccl,_me))
      continue;
    for(it = pEnt->sa.replicas.begin();it != pEnt->sa.replicas.end();++it){
      accum_set.insert(*it);
    }
  }
  assert(pthread_mutex_unlock(&_slices_mutex) == 0);
  // Weed out passive hosts from sitemap
  assert(pthread_mutex_lock(&_sites_mutex) == 0);
  sitemap::iterator siteiter;
  std::set<host>::iterator sit;
  for(siteiter = _sites.begin();siteiter != _sites.end(); ++ siteiter){
    std::vector<host> to_erase;
    for(sit = siteiter->second.begin();sit != siteiter->second.end(); ++sit){
      if(accum_set.count(*sit) != 1){
        to_erase.push_back(*sit);
        wfs_log(WFS_DBG_3,"slice_manager::get_active_set: Removed passive host:%s from site map.\n",
              (*sit).str().c_str());
      }
    }
    for(unsigned i = 0; i < to_erase.size(); i++){
      siteiter->second.erase(to_erase[i]);
    }
  }
  assert(pthread_mutex_unlock(&_sites_mutex) == 0);
  return SM_SUCCESS;
}

/*
 * While it would be more efficient to get the list from sitemap, we can't tell from that structure
 * if the related slice is locked or even exists. Use the opportunity to weed out inactive hosts from
 * the sitemap.
 */
slice_manager::ret_code
slice_manager::get_all_active_hosts(std::vector<host> & hosts)
{
  std::set<host> accum_set;
  slice_manager::ret_code ret = get_active_set(accum_set);
  assert (ret == slice_manager::SM_SUCCESS);

  hosts.assign(accum_set.begin(),accum_set.end());

  wfs_log(WFS_DBG_3,"slice_manager::get_all_active_hosts: Returned %u active hosts\n",
	   (unsigned)hosts.size());

  return SM_SUCCESS;
}

// Not very efficient, but the best we can do with the current data structures.
slice_manager::ret_code
slice_manager::get_maintenence_slices_for_host(const host h,std::vector<wfs_slice_id>& slicev)
{

  // can't call on a non-server
  if (_me.addr == 0) {
    return SM_GENERAL_ERR;
  }

  slicemap::const_iterator iter;
  std::list<host>::const_iterator  it;
  const slice_ent * pEnt;
  assert(pthread_mutex_lock(&_slices_mutex) == 0);
  for(iter=_slices.begin(); iter != _slices.end(); ++iter){
    pEnt = iter->second;std::set<host> accum_set;
    for(it = pEnt->sa.replicas.begin();it != pEnt->sa.replicas.end();++it){
      if(*it == h){
        slicev.push_back(pEnt->sa.slice.val());
        break;
      }
    }
  }
  assert(pthread_mutex_unlock(&_slices_mutex) == 0);

  wfs_log(WFS_DBG_3,"slice_manager::get_maintenence_slices_for_host: Host:%s is part of %u slices\n",
	   h.str().c_str(),(unsigned)slicev.size());

  return SM_SUCCESS;
}

// FOR DEBUGGING
void
slice_manager::print_all_known_slices()
{
  slicemap::iterator iter;
  std::list<host>::iterator  it;
  slice_ent * pEnt;
  assert(pthread_mutex_lock(&_slices_mutex) == 0);
  printf("---PRINTING ALL KNOWN SLICES---\n");
  for(iter=_slices.begin(); iter != _slices.end(); ++iter){
    pEnt = iter->second;std::set<host> accum_set;
    printf("SLICE:%016"PRIx64"\n",pEnt->sa.slice.val());
    printf("\tACQUIRE STATE:%s\n",pEnt->locked?"LOCKED":"FREE");
    printf("\tPRIMARY:%s%s on SITE:%s\n",
        (static_cast<std::string>(pEnt->primary())).c_str()
        ,(pEnt->amiprimary(_me)?"(THIS HOST)":""),pEnt->primary().site_id.c_str());
    printf("\tSTATE ON SERVER:%s\n",pEnt->lock_state.c_str());
    printf("\tREPLICATION LIST:\n");
    int i = 1;
    for(it = pEnt->sa.replicas.begin();it != pEnt->sa.replicas.end();++it,++i){
      printf("\t\tREP%d: %s on site:%s\n",i,(static_cast<std::string>(*it)).c_str(),it->site_id.c_str());
    }
    does_entry_comply(pEnt);
    printf("\tCOMPLIES WITH POLICY? %s\n",(pEnt->compliance == slice_ent::OK)?"YES":"NO");
    printf("\tPROBLEMS:%016"PRIx64"\n",pEnt->repair_items);

  }
  assert(pthread_mutex_unlock(&_slices_mutex) == 0);
}

// Need to hold _slice_mutex for this. Don't call for slices that are being sync'ed.
bool
slice_manager::does_entry_comply(slice_ent * pEnt, bool dontCheckWhenSyncing)
{
  // Only Check Locked non-syncing slices.
  if(!pEnt->islocked(_ccl,_me) || (dontCheckWhenSyncing && pEnt->compliance == slice_ent::SYNCING))
    return true;

  if(pEnt->compliance != slice_ent::SYNCING)
    pEnt->compliance = slice_ent::OK; // Assume it is OK and check for problems.
  wfs_slice_id sid = pEnt->sa.slice.val();

  wfs_log(WFS_DBG_3,"slice_manager::does_entry_comply: Checking Slice:%016"PRIx64"\n",sid);

  // Check site
  if(pEnt->amiprimary(_me) &&  _me.site_id != pEnt->sa.policy.prim_site){

    wfs_log(WFS_DBG_2,"slice_manager::does_entry_comply: Slice:%016"PRIx64" Primary site:%s Should be %s\n"
           ,sid,pEnt->primary().site_id.c_str(),pEnt->sa.policy.prim_site.c_str());

    pEnt->set_errstat(slice_ent::WRONG_PRIMARY_SITE);
    pEnt->compliance = slice_ent::REPAIR;
  }
  if(pEnt->sa.replicas.size() < pEnt->sa.policy.rep_num){

    wfs_log(WFS_DBG_2,"slice_manager::does_entry_comply: Slice:%016"PRIx64" Has only %u replicas. Needs %d\n"
	      ,sid,(unsigned)pEnt->sa.replicas.size(),pEnt->sa.policy.rep_num);

    pEnt->set_errstat(slice_ent::NOT_ENOUGH_HOSTS);
    pEnt->compliance = slice_ent::REPAIR;
  }
  std::set<std::string> knownsites;
  for(std::list<host>::const_iterator it = pEnt->sa.replicas.begin();
      it != pEnt->sa.replicas.end();++it){
    knownsites.insert(it->site_id);
  }
  if(knownsites.size() < pEnt->sa.policy.site_num){

    wfs_log(WFS_DBG_2,"slice_manager::does_entry_comply: Slice:%016"PRIx64" Has only %u sites. Needs %d\n"
	      ,sid,(unsigned)knownsites.size(),pEnt->sa.policy.site_num);

    pEnt->set_errstat(slice_ent::NOT_ENOUGH_SITES);
    pEnt->compliance = slice_ent::REPAIR;
  }
  if(pEnt->compliance == slice_ent::REPAIR){

    wfs_log(WFS_DBG_2,"slice_manager::does_entry_comply: Slice:%016"PRIx64" Marked for repair.\n",sid);

    return false;
  } else {

    wfs_log(WFS_DBG_3,"slice_manager::does_entry_comply: Slice:%016"PRIx64" complies with policy\n",sid);

    pEnt->compliance = slice_ent::OK;
    pEnt->reset_all_errstats();
    return true;
  }
}

inline void
slice_manager::find_hosts_in_site(const wfs_slice_id sidx, const std::string sitename,
    int & needed_hostsx,std::set<host> hostsetx,
    std::list<std::pair<wfs_slice_id,host> >& recruitsx,
    slice_manager::ret_code& retx)
{
  sitemap::const_iterator iterx;
  std::set<host>::const_iterator set_itx;
  iterx = _sites.find(sitename);
  if(iterx == _sites.end()){
    return;
  }
  for(set_itx = iterx->second.begin();set_itx != iterx->second.end() && needed_hostsx > 0;++set_itx){
    if(hostsetx.count(*set_itx)== 0){
      // Found one. add it.
      recruitsx.push_back(std::make_pair(sidx,*set_itx));
      needed_hostsx--;
      // Since a host can only appear once in any site, we don't need to add it to siteset
      hostsetx.insert(*set_itx);
      retx = SM_SUCCESS;

      wfs_log(WFS_DBG_3,"slice_manager::suggest_repair:Adding Host:%s (Site:%s) to Slice:%016"PRIx64".\n",
            (*set_itx).str().c_str(),set_itx->site_id.c_str(),sidx);

    }
  }
}

/*
 * Repair Logic:
 * ENOUGH_HOSTS && ENOUGH_SITES - Nothing to do.
 * NOT_ENOUGH_HOSTS && ENOUGH_SITES - Add any new host from any site.
 * ENOUGH_HOSTS && NOT_ENOUGH_SITES - Need to move a replica from a host in existing site to a host in a new site.
 *   for this call this is exactly as the case below. The difference will be in report_sync_complete.
 * NOT_ENOUGH_HOSTS && NOT_ENOUGH_SITES - Add a new host from a new site.
 */
slice_manager::ret_code
slice_manager::suggest_repair(slice_ent * pEnt,std::list<std::pair<wfs_slice_id,host> > & recruits,
    const std::set<host> active_set)
{
  wfs_slice_id sid = pEnt->sa.slice.val();
  slice_manager::ret_code ret = SM_CANT_REPAIR;
  sitemap::const_iterator iter;
  std::set<host>::const_iterator set_it;

  if(pEnt->compliance !=  slice_ent::REPAIR)
    return SM_SUCCESS;

  std::set<host> hostset;
  std::set<std::string> siteset;
  std::vector<std::string> sites_vec;
  for(std::list<host>::const_iterator it = pEnt->sa.replicas.begin();
      it != pEnt->sa.replicas.end();++it){
    if((siteset.insert(it->site_id)).second){
      sites_vec.push_back(it->site_id);
    }
    hostset.insert(*it);
  }
  int needed_sites = pEnt->sa.policy.site_num - siteset.size();
  int needed_hosts = pEnt->sa.policy.rep_num - hostset.size();

  /*
   * If the primary site is at the wrong site. First look for an active primary site in the old replicas set.
   * If can't find, try and find an active host in the sitemap
   */
  bool foundold = false;
  if(pEnt->is_errstat(slice_ent::WRONG_PRIMARY_SITE)){
    for(set_it = pEnt->old_replicas.begin();set_it != pEnt->old_replicas.end();++set_it){
      if(active_set.count(*set_it) == 1 &&
          set_it->site_id == pEnt->sa.policy.prim_site &&
          hostset.count(*set_it) != 1){
        // Bingo
        recruits.push_back(std::make_pair(sid,*set_it));
        needed_hosts--;
        foundold = true;
        ret = SM_SUCCESS;

        wfs_log(WFS_DBG_3,"slice_manager::suggest_repair:Adding previously used replica :%s (Site:%s) to Slice:%016"PRIx64".\n",
              (*set_it).str().c_str(),set_it->site_id.c_str(),pEnt->sa.slice.val());

        break;
      }
    }
    if(!foundold){
      assert(pthread_mutex_lock(&_sites_mutex) == 0);
      find_hosts_in_site(sid,pEnt->sa.policy.prim_site,needed_hosts,hostset,recruits,ret);
      assert(pthread_mutex_unlock(&_sites_mutex) == 0);
    }
    return ret;
  }

  assert(pthread_mutex_lock(&_sites_mutex) == 0);
  if(needed_sites == 0 && _me.site_id != ""){
    // We can pick from any site. Try and get from the primary site.
    find_hosts_in_site(sid,_me.site_id,needed_hosts,hostset,recruits,ret);
  }
  // Get hosts from other sites.


  int rep_level = pEnt->sa.policy.rep_num;
  int rep_sites = pEnt->sa.policy.site_num;
  std::map<std::string,std::pair<int,int> > nodes_per_site;
  for(sitemap::const_iterator it = _sites.begin(); it != _sites.end();++it){
    nodes_per_site[it->first] = std::pair<int,int>(0,it->second.size());
  }
  // This host is never in _sites. Need to count it in capacity too.
  if(nodes_per_site.find(_me.site_id) != nodes_per_site.end())
    nodes_per_site[_me.site_id].second++;
  assert(nodes_per_site.size() == _sites.size());
  for(std::list<host>::const_iterator it = pEnt->sa.replicas.begin();
  it != pEnt->sa.replicas.end();++it){
    (nodes_per_site[it->site_id].first)++;
  }

  host h;
  std::string s;

  // need to recruit the maximum of needed_sites and needed_hosts
  int needed = (needed_hosts>needed_sites)?needed_hosts:needed_sites;

  while(needed > 0){
    ret = get_random_replica(rep_level-needed,rep_sites,siteset,
			     sites_vec,hostset,nodes_per_site,h,s);
    if(ret != SM_SUCCESS){
      break;
    }

    // only use this host if we need a host, or if it gives us a new site
    if( needed_hosts == 0 && siteset.find(s) != siteset.end() ) {
      ret = SM_NOENT;
      break;
    }

    hostset.insert(h); //  an unused node is always returned on success
    recruits.push_back(std::make_pair(sid,h));
    (nodes_per_site[h.site_id].first)++;
    needed_hosts--;
    if((siteset.insert(s)).second){
      // The site was inserted
      sites_vec.push_back(s);
      needed_sites--;
    }
    needed--;
  }

  assert(pthread_mutex_unlock(&_sites_mutex) == 0);

  return ret;
}

slice_manager::ret_code
slice_manager::check_all_slices(std::list<std::pair<wfs_slice_id,host> > & recruits)
{
  slice_manager::ret_code ret = SM_SUCCESS;
  slicemap::iterator iter;
  slice_ent * pEnt;
  std::set<host> active_set;

  // can't call on a non-server
  if (_me.addr == 0) {
    return SM_GENERAL_ERR;
  }

  sync_state();
  // This will update the sitemap and give a set of active hosts.
  get_active_set(active_set);

  assert(pthread_mutex_lock(&_slices_mutex) == 0);
  for(iter = _slices.begin(); iter != _slices.end();++iter){
    pEnt = iter->second;
    // Only repair entries for which we are primaries.
    if(pEnt->amiprimary(_me) && pEnt->islocked(_ccl,_me) && !does_entry_comply(pEnt)){
      ret = suggest_repair(pEnt,recruits,active_set);
      if(ret != SM_SUCCESS){

        wfs_log(WFS_DBG_2,"slice_manager::check_all_slices: Slice:%016"PRIx64". Assignment can't be repaired at the moment.\n"
              ,pEnt->sa.slice.val());

        continue;
      }
      pEnt->compliance = slice_ent::SYNCING;
    }
  }
  assert(pthread_mutex_unlock(&_slices_mutex) == 0);
  return SM_SUCCESS;
}

slice_manager::ret_code
slice_manager::report_sync_complete(const wfs_slice_id sid,const host h,std::vector<host>& dropped,
    bool& do_lock_rpc)
{
  slice_ent * pEnt;
  slice_manager::ret_code ret = SM_SUCCESS;
  slicemap::iterator iter;
  // Save the last state
  std::string old_state;
  // Only replace assignment if locking was successful.
  slice_assignment newsa("",0);
  int r;
  do_lock_rpc = false;

  // can't call on a non-server
  if (_me.addr == 0) {
    return SM_GENERAL_ERR;
  }

  assert(pthread_mutex_lock(&_slices_mutex) == 0);
  iter = _slices.find(sid);
  if(iter == _slices.end()){

    wfs_log(WFS_DBG_1, "report_sync_complete: %016"PRIx64" no such slice!\n", sid );

    ret = SM_NOENT;
    goto exit;
  }
  pEnt = iter->second;
  // Make sure that there is something to fix
  if(does_entry_comply(pEnt,false)){
    pEnt->compliance = slice_ent::OK;
    ret = SM_SUCCESS;
    goto exit;
  } else {
    pEnt->compliance = slice_ent::REPAIR;
  }

  if(pEnt->is_errstat(slice_ent::WRONG_PRIMARY_SITE) && (pEnt->sa.policy.prim_site == h.site_id) ){
    //If the synced host is from a "good" site release this lock with the server and hint other host to lock it.
    if( (r = _ccl->release(&pEnt->sa)) == wcc_protocol::OK){

      wfs_log(WFS_DBG_2,"report_sync_complete:Released lock for slice %016"PRIx64". Hoping host:%s will lock it.\n",
            sid,h.str().c_str());

      pEnt->locked = false;
      do_lock_rpc = true;
      ret = SM_SUCCESS;
    } else {

      wfs_log(WFS_DBG_1,"report_sync_complete: %016"PRIx64" couldn't be released (%d)!\n", sid, r );

      ret = SM_RSM_ERR;
    }
    goto exit;
  }
  old_state = pEnt->sa.to_canonical_string();
  newsa = pEnt->sa;

  if(pEnt->is_errstat(slice_ent::NOT_ENOUGH_SITES) && !pEnt->is_errstat(slice_ent::NOT_ENOUGH_HOSTS)){
    // This means we need to remove one replica.
    dropped.push_back(newsa.replicas.back());
    pEnt->old_replicas.insert(newsa.replicas.back()); // Remember this was once a replica
    newsa.replicas.pop_back();
  }
  // Note that we don't verify that the new node is from a "correct" site. We might not
  // even be solving the problem (will be checked below).
  newsa.replicas.push_back(h); // Add the new replica.
  // Try and renew the slice.
  if( (r = _ccl->renew_slice(&newsa)) != wcc_protocol::OK){

    wfs_log(WFS_DBG_1, "report_sync_complete: %016"PRIx64" couldn't renew (%d)!\n", sid, r );

    ret = SM_RSM_ERR;
    goto exit;
  }
  pEnt->sa = newsa;
  pEnt->lock_state = pEnt->sa.to_canonical_string();

  if(does_entry_comply(pEnt,false)){
    pEnt->compliance = slice_ent::OK;
  } else {
    pEnt->compliance = slice_ent::REPAIR;
  }

exit:
  assert(pthread_mutex_unlock(&_slices_mutex) == 0);
  return ret;
}

slice_manager::ret_code
slice_manager::report_sync_complete(const wfs_slice_id sid,const host h,std::vector<host>& dropped)
{
  slice_manager::ret_code ret;
  bool do_lock_rpc = false;
  ret = report_sync_complete(sid,h,dropped,do_lock_rpc);
  if(do_lock_rpc){
    wfs_log(WFS_DBG_1,"XXXX:report_sync_complete: Need to tell %s to lock %016"PRIx64" but return code will be lost. Use overloaded version!\n"
        , h.str().c_str(), sid);
  }
  return ret;
}

slice_manager::ret_code
slice_manager::sync_aborted(const wfs_slice_id sid,const host h)
{
  slice_ent * pEnt;
  slice_manager::ret_code ret = SM_SUCCESS;
  slicemap::iterator iter;

  // can't call on a non-server
  if (_me.addr == 0) {
    return SM_GENERAL_ERR;
  }

  assert(pthread_mutex_lock(&_slices_mutex) == 0);
  iter = _slices.find(sid);
  if(iter == _slices.end()){
    ret = SM_NOENT;
  } else {
    pEnt = iter->second;
    pEnt->compliance = slice_ent::REPAIR;
  }
  assert(pthread_mutex_unlock(&_slices_mutex) == 0);

  return ret;
}


template<class T>
static typename std::list<T>::iterator
get_host_position(std::list<T>& lst,const T h, int& offset)
{
  offset = 0;
  typename std::list<T>::iterator it;
  for(it = lst.begin(); it != lst.end(); ++it,offset++){
    if(h == *it)
      return it;
  }
  offset = -1;
  return it;
}

/*
 * report_host_down(h) {
 * slice set s = set of slices where h is an active replica
 * if h has no unexpired locks in s {
 *  foreach S in s {
 *      if the local host is the current or new primary for S {
 *         remove h from the active replicaset of S
 *        add h to the inactive replicaset of S
 *         lock S using the new replicaset
 *        set the status of S to REPAIR
 *     }
 * }
 *  return SM_SUCCESS
 * } else {
 *  return SM_CANT_REPAIR
 * }
 *}
 */
slice_manager::ret_code
slice_manager::report_host_down(const host h)
{
  slicemap::iterator iter;
  std::list<host>::iterator it;
  slice_ent * pEnt;
  int offset = 0,rsmret;
  slice_assignment tempSa("",0);;
  slice_manager::ret_code ret = SM_CANT_REPAIR;

  // can't call on a non-server
  if (_me.addr == 0) {
    return SM_GENERAL_ERR;
  }

  if(h == _me){
    return SM_CANT_REPAIR;
  }

  wfs_log(WFS_DBG_1,"slice_manager::report_host_down: Called for host:%s\n"
          ,h.str().c_str());

  sync_state();
  assert(pthread_mutex_lock(&_slices_mutex) == 0);
  for(iter=_slices.begin(); iter != _slices.end(); ++iter){
    bool needslock = false;
    pEnt = iter->second;
    if(pEnt->amiprimary(_me) &&
        (it = get_host_position<host>(pEnt->sa.replicas,h,offset)) != pEnt->sa.replicas.end() ){
      // this host is the primary and h is one of the replicas.
      tempSa = pEnt->sa;
      tempSa.replicas.remove(h);
    } else if(!pEnt->islocked(_ccl,_me) &&
             (pEnt->primary() == h) &&
             (it = get_host_position<host>(pEnt->sa.replicas,_me,offset)) != pEnt->sa.replicas.end() &&
             (offset == 1) ){
      // h is the current primary and this host is second on the replica list
      tempSa = pEnt->sa;
      tempSa.replicas.pop_front(); // Remove the old primary.
      needslock = true;
    } else {
      continue;
    }
    //Try and lock/renew the new slice.
    if (needslock) {
      rsmret = _ccl->acquire_slice(&tempSa,_lease_period,true);

       wfs_log(WFS_DBG_1,"slice_manager::report_host_down: Locking Slice:%016"PRIx64" as the new primary. RSM Return code:%d\n"
              ,tempSa.slice.val(),rsmret);

    } else {
      rsmret = _ccl->renew_slice(&tempSa);

       wfs_log(WFS_DBG_2,"slice_manager::report_host_down: Slice:%016"PRIx64" Removed second . RSM Return code:%d\n"
              ,tempSa.slice.val(),rsmret);
    }
    if(rsmret != wcc_protocol::OK){
      wfs_log(WFS_DBG_1,"slice_manager::report_host_down: Slice:%016"PRIx64". Assignment can't be repaired. Can't be locked with server. Error:%d\n"
            ,tempSa.slice.val(),rsmret);
      continue;
    }

    // Update the new Entry
    pEnt->sa = tempSa;
    pEnt->lock_state = pEnt->sa.to_canonical_string();
    pEnt->locked = true;
    // Remember the down host
    pEnt->old_replicas.insert(h);
    pEnt->compliance = slice_ent::REPAIR;

    ret = SM_SUCCESS; // Fixed at least one slice.
  }
  assert(pthread_mutex_unlock(&_slices_mutex) == 0);

  return ret;
}

slice_manager::ret_code
slice_manager::get_policy_for_slice(const wfs_slice_id sid,rep_policy& pol)
{
  slice_manager::ret_code ret = SM_SUCCESS;
  slice_ent * pEnt;
  slicemap::const_iterator iter;
  assert(pthread_mutex_lock(&_slices_mutex) == 0);
  if((iter = _slices.find( wfs_slice::get_slice_id(sid))) == _slices.end()){
    ret = SM_NOENT;
  } else {
    pEnt = iter->second;
    pol = pEnt->sa.policy;
  }
  assert(pthread_mutex_unlock(&_slices_mutex) == 0);
  return ret;
}

slice_manager::ret_code
slice_manager::release_slice(const wfs_slice_id sid)
{
  slice_manager::ret_code ret = SM_SUCCESS;
  slice_ent * pEnt;

  // can't call on a non-server
  if (_me.addr == 0) {
    return SM_GENERAL_ERR;
  }

  slicemap::iterator iter;
  assert(pthread_mutex_lock(&_slices_mutex) == 0);
  if((iter = _slices.find( wfs_slice::get_slice_id(sid))) == _slices.end()){
    ret = SM_NOENT;
  } else {
    pEnt = iter->second;
    _ccl->release(&(pEnt->sa));
  }
  assert(pthread_mutex_unlock(&_slices_mutex) == 0);
  return ret;
}

wfs_file_id slice_manager::slice_ent::get_fid()
{
  // the last four bits of random() are known to be not-so-random, so
  // shift them off.
  wfs_file_id rand_id = (random() >> 4);
  rand_id <<= 28;
  rand_id += (random() >> 4);
  rand_id &= ~wfs_slice::SLICE_MASK;
  rand_id |= sa.slice.val();

  wfs_log(WFS_DBG_3,"Created new fid %016"PRIx64" in slice %016"PRIx64"\n", rand_id, sa.slice.val() );

  count++;
  return rand_id;
}

