/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang,
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


// RPC stubs for overfuse to talk to overserve.

#include "wfs_client.h"

#include <arpa/inet.h>
#include <assert.h>
#include <chash.h>
#include <host_marshall.h>
#include <rpc.h>
#include <set>
#include <sys/stat.h>
#include <unistd.h>

#include <wfs_file.h>
#include <wfs_misc.h>

wfs_client::wfs_client( wfs_node me, slice_manager *slices, vivaldi *viv,
			wfs_client_manager *clients )
  : _slices(slices), _me(me), _viv(viv), _clients(clients),
    _local_server(NULL) {
}

bool wfs_client::isfile( wfs_file_handle fh ) {

  return wfs_isfile(fh);

}

bool wfs_client::isdir( wfs_file_handle fh ) {

  return wfs_isdir(fh);

}

bool wfs_client::islink( wfs_file_handle fh ) {

  return wfs_islink(fh);

}

template<class R> wfs_protocol::status
wfs_client::repeated_call( wfs_file_handle fh,
			   unsigned int proc,
			   marshall &req,
			   R & r ) {

  wfs_protocol::status status = wfs_protocol::OK;
  unsigned i = 1;
  bool repeat = false;
  int rep_level;
  wfs_node file_node;
  std::vector<host> reps;
  slice_manager::ret_code r2 =
    _slices->get_slice_replicaset( wfs_slice::get_slice_id(fh.fid),
				   reps, rep_level );
  assert (r2 == slice_manager::SM_SUCCESS);

  file_node = make_node(reps[i-1]);

  do {

    repeat = false;
    wfs_file_id node_id_short = (wfs_file_id) file_node.id;

    // if we've had a timeout to this node "recently", then just skip it
    // straight away
    if( fh.get_cue(WFS_CUE_CONSISTENCY).val == WFS_CUE_EVENTUAL &&
	i < (unsigned)rep_level && i < reps.size() &&
	(_last_timeouts.find(node_id_short) !=
	 _last_timeouts.end())  &&
	(_last_timeouts[node_id_short] >
	 time(NULL) - WFS_CLIENT_TIMEOUT_CACHE_SEC) ) {

      wfs_log(WFS_DBG_1, "Skipping calling proc %d on node %016"PRIx64" due to cached timeout\n",
		proc, file_node.id );

      i++;
      file_node = make_node(reps[i-1]);
      repeat = true;
      continue;

    }

    _last_timeouts.erase( node_id_short );
    //marshall m = req;
    // since we might have to reuse it, don't let rpc delete the buffer
    req.dont_del_buf();

    wfs_log(WFS_DBG_2, "Calling proc %x on node %016"PRIx64"\n", proc, file_node.id );

    rpcc *client = _clients->get_client( file_node, fhto(fh) );
    if( client == NULL ) {
      wfs_log(WFS_DBG_2, "Binding failed on node %016"PRIx64"\n", file_node.id );
      status = rpc_const::timeout_failure;
    } else {
      status = client->call_m( proc, req, r, fhto(fh), "host" );
      _clients->done_with_client(file_node);
      wfs_log(WFS_DBG_2, "Finished proc %x on node %016"PRIx64", status %d\n", proc,
		file_node.id, status );
    }

    if( status == wfs_protocol::NOTMASTER ) {
      _slices->sync_state();
    }

    // don't worry about the lax cue here, the server will check it
    // when responding as appropriate
    if( (status == rpc_const::timeout_failure ||
	 status == wfs_protocol::NOTMASTER) ) {

      if( status == rpc_const::timeout_failure ) {
	_last_timeouts[node_id_short] = time(NULL);
      }

      // if the node has changed in our current position, do-over
      std::vector<host> reps2;
      int rl2;
      _slices->get_slice_replicaset( wfs_slice::get_slice_id(fh.fid), reps2,
				     rl2 );
      if( (reps2.size() < i && reps2.size() > 0) || 
	  (reps2.size() >= i && reps2[i-1] != reps[i-1]) ) {

	i = 1;
	file_node = make_node(reps2[i-1]);
	reps = reps2;
	rep_level = rl2;
	repeat = true;

	wfs_log(WFS_DBG_1, "Repeat-calling proc %x on node %016"PRIx64" due to changed replica set\n",
		  proc, file_node.id );

      } else if( i < (unsigned)rep_level && i < reps.size() ) {

        i++;
        file_node = make_node(reps[i-1]);
        repeat = true;

        wfs_log(WFS_DBG_1, "Repeat-calling proc %x on node %016"PRIx64" due to timeout\n",
            proc, file_node.id );

      }

    } else if( status == wfs_protocol::STALE ) {
      // TODO: limit time or iterations
      repeat = true;

      wfs_log(WFS_DBG_1, "Repeat-calling proc %d on node %016"PRIx64" due to staleness\n",
		proc, file_node.id );
    }

  } while( repeat );

  return status;

}

wfs_protocol::status
wfs_client::getfile( wfs_file_handle &fh,
		     const std::string &user,
		     wfs_protocol::fileinfo &fin,
		     host *client ) {

  wfs_node file_node = get_primary( file_handle_to_file_id( fh ) );
  if( file_node.id == 0 ) {
    return wfs_protocol::IOERR;
  }

  host h;
  if( client != NULL ) {
    h = *client;
  }

  wfs_protocol::status status;

  if( is_local(file_node) ) {
    status = _local_server->getfile(user, fh, h, fin );
  } else {
    marshall m;
    m << user;
    m << fh;
    m << h;
    status = repeated_call( fh, wfs_protocol::getfile, m, fin );
  }

  if( status != wfs_protocol::OK ) {
    return status;
  }

  // version files don't contain the regular contents, so the size should
  // be however big the ASCII equivalent of an int is.
  if( fh.is_version ) {
    fin.size = WFS_VERSION_FILE_SIZE;
  }

  if( fh.version.version_num >= 0 && fh.version != fin.version ) {
    wfs_log(WFS_DBG_1, "BADVERSION caught on client!\n");
    return wfs_protocol::BADVERSION;
  }

  // Check for permanent cues, since they might not yet be set in the
  // file handle
  fh.set_perm_cues( fin.perm_cues );
  return status;

}

//Get info for files, files *must* all belong to one primary server
wfs_protocol::status
wfs_client::getfiles( std::vector<wfs_file_handle> &fhs,
		      const std::string &user,
		      std::vector<wfs_protocol::fileinfo> &fins,
		      host *client ) {

  wfs_node file_node = get_primary( file_handle_to_file_id( fhs[0] ) );
  if( file_node.id == 0 ) {
    return wfs_protocol::IOERR;
  }

  host h;
  if( client != NULL ) {
    h = *client;
  }

  wfs_protocol::status status;

  if( is_local(file_node) ) {
    status = _local_server->getfiles(user, fhs, h, fins );
  } else {
    marshall m;
    m << user;
    m << fhs.size();
    for( std::vector<wfs_file_handle>::iterator i = fhs.begin(); i != fhs.end(); i++) {
      m << *i;
    }
    m << h;
    status = repeated_call( fhs[0], wfs_protocol::getfiles, m, fins );
  }

  if( (status != wfs_protocol::OK) || (fhs.size() != fins.size()) ) {
    return status;
  }

  for( int i = 0; i < fins.size(); i++) {
    wfs_file_handle fh = fhs[i];
    wfs_protocol::fileinfo fin = fins[i];
    // version files don't contain the regular contents, so the size should
    // be however big the ASCII equivalent of an int is.
    if( fh.is_version ) {
      fin.size = WFS_VERSION_FILE_SIZE;
    }

    if( fh.version.version_num >= 0 && fh.version != fin.version ) {
      wfs_log(WFS_DBG_1, "BADVERSION caught on client!\n");
      return wfs_protocol::BADVERSION;
    }

    // Check for permanent cues, since they might not yet be set in the
    // file handle
    fh.set_perm_cues( fin.perm_cues );

  }
  return status;
}

wfs_protocol::status wfs_client::getdir( wfs_file_handle &fh,
					 const std::string &user,
					 wfs_protocol::dirinfo &fin,
					 host *client ) {

  wfs_node dir_node = get_primary( file_handle_to_file_id( fh ) );
  if( dir_node.id == 0 ) {
    return wfs_protocol::IOERR;
  }

  wfs_protocol::status status;
  host h;
  if( client != NULL ) {
    h = *client;
  }

  if( is_local(dir_node) ) {
    status = _local_server->getdir( user, fh, h, fin );
  } else {
    marshall m;
    m << user;
    m << fh;
    m << h;
    status = repeated_call( fh, wfs_protocol::getdir, m, fin );
  }

  if( status != wfs_protocol::OK ) {
    // if this is lax mode, let's just make the directory up.  We have the
    // file id, we know it exists somewhere, it just hasn't been repaired
    // to the right places yet.  We'll just let the client start using it
    // like an empty directory and hope everything works out
    if( fh.get_cue(WFS_CUE_CONSISTENCY).val == WFS_CUE_EVENTUAL &&
	(status == -1 || status == wfs_protocol::NOENT) ) {
      fin.perm_cues = fh.perm_cues_string();
      fin.permissions.clear();
      fin.permissions[ANYONE_IDENT] = WFS_PERMISSION_ALL;
      return wfs_protocol::OK;
    } else {
      return status;
    }
  }

  // Check for permanent cues, since they might not yet be set in the
  // file handle
  fh.set_perm_cues( fin.perm_cues );
  return status;

}

wfs_protocol::status wfs_client::lookup( wfs_file_handle dir_handle,
					 std::string name,
					 const std::string &user,
					 wfs_file_handle &xfh,
					 host *client ) {

  wfs_protocol::lookupinfo li;
  wfs_protocol::status s =  lookup( dir_handle, name, user, li, client );
  xfh = li.fh;
  return s;

}

wfs_protocol::status wfs_client::lookup( wfs_file_handle dir_handle,
					 std::string name,
					 const std::string &user,
					 wfs_protocol::lookupinfo &li,
					 host *client ) {

  if( dir_handle.is_nodes ) {
    int oct_1;
    int oct_2;
    int oct_3;
    int oct_4;
    int port;
    int n = sscanf( name.c_str(), "%d.%d.%d.%d:%d", &oct_1, &oct_2, &oct_3,
		    &oct_4, &port );
    if( n != 5 ) {
      return wfs_protocol::NOENT;
    }
    li.fh = dir_handle;
    li.fh.is_nodes = false;
    li.fh.site = "";
    char ip[100];
    sprintf( ip, "%d.%d.%d.%d", oct_1, oct_2, oct_3, oct_4 );
    li.fh.node = make_node( ip, port );
    return wfs_protocol::OK;
  }

  if( dir_handle.is_sites ) {
    li.fh = dir_handle;
    li.fh.is_sites = false;
    li.fh.site = name;
    return wfs_protocol::OK;
  }

  // we don't need to go all the way to the server to set cues, you know
  li.fh = dir_handle; // copies the cues
  li.fh.inc_depth();
  wfs_node dir_node;
  int x;
  if( (x = search_and_set_cues( name, li.fh )) ) {

    wfs_log(WFS_DBG_1, "Setting cue on client side\n" );

    // if this is the debug cue and should get passed along to the home node
    if( x == WFS_CUE_DEBUG ) {
      dir_node = _me;
    } else {
      return wfs_protocol::OK;
    }

  } else {
    dir_node = get_primary( file_handle_to_file_id( dir_handle ) );
    if( dir_node.id == 0 ) {
      return wfs_protocol::IOERR;
    }
  }

  host h;
  if( client != NULL ) {
    h = *client;
  }

  wfs_protocol::status status;

  if( is_local(dir_node) ) {
    status = _local_server->lookup( user, dir_handle, name, h, li );
  } else {
    marshall m;
    m << user;
    m << dir_handle;
    m << name;
    m << h;
    status = repeated_call( dir_handle, wfs_protocol::lookup, m, li );
  }

  // it's important that the node field propagates along with the handle
  // for directories
  li.fh.node = dir_handle.node;
  li.fh.version = dir_handle.version;
  return status;

}

bool is_meta_handle( const wfs_file_handle &fh ) {
  return ( fh.is_nodes || fh.is_sites );
}

wfs_protocol::status wfs_client::create( wfs_file_handle dir_handle,
					 std::string name,
					 const std::map<std::string,
					 wfs_permission> &permissions,
					 wfs_file_type type,
					 const std::string &user,
					 wfs_file_handle &new_fh,
					 host *client ) {

  // don't allow creates to happen in the .nodes directory
  if( dir_handle.is_nodes || dir_handle.is_sites ) {
    return wfs_protocol::IOERR;
  }

  // TODO: rewrite this method quite a bit

  // Create a policy object based on the site, replevel, and numsites cues
  // ask the slice manager for a new fid in a slice fitting that policy
  //   if successful, contact the primary to create the file
  //   if TRYHOST, then contact the host, tell it to create the slice AND
  //      the file
  // then create the directory entry on the directory primary

  std::string site;
  if( dir_handle.site == WFS_RANDOM_SITE ) {

    std::vector<host> hosts;
    _slices->get_all_active_hosts(hosts);
    std::set<std::string> site_map;
    std::vector<std::string> site_vec;
    for( unsigned int i = 0; i < hosts.size(); i++ ) {
      if( hosts[i].site_id != "" ) {
	if( site_map.insert(hosts[i].site_id).second ) {
	  site_vec.push_back(hosts[i].site_id);
	}
      }
    }
    int r = random() % site_vec.size();
    site = site_vec[r];

  } else if( dir_handle.site != "" ) {
    site = dir_handle.site;
  } else if( dir_handle.node.site_id != "" ) {
    site = dir_handle.node.site_id;
  } else {
    site = _me.site_id;
  }

  rep_policy pol( fhreplevel(dir_handle), rep_policy::STRICT, site,
		  fhrepsites(dir_handle) );

  if( dir_handle.get_cue(WFS_CUE_DIR_LOC_POLICY).val ==
      WFS_CUE_KEEP_TOGETHER ) {
    _slices->get_policy_for_slice( wfs_slice::get_slice_id(dir_handle.fid),
				   pol );
    // use the desired replevel and repsite; just use the parent
    // policy for the primary site
    pol.rep_num = fhreplevel(dir_handle);
    pol.site_num = fhrepsites(dir_handle);
  }

  wfs_log(WFS_DBG_1, "Creating %s in site %s in dir %s\n", name.c_str(),
	    pol.prim_site.c_str(), dir_handle.print().c_str() );

  wfs_node dir_node = get_primary( file_handle_to_file_id( dir_handle ) );
  if( dir_node.id == 0 ) {
    return wfs_protocol::IOERR;
  }

  wfs_protocol::status s;
  wfs_file_handle final_fh;
  do {

    wfs_file_id new_id;
    wfs_node file_node;

    // if I don't have a local server, then I can't call get_new_id if
    // it might create a slice for me.  So fake it.
    if( _local_server == NULL && pol.prim_site == _me.site_id ) {

      file_node = _me;
      new_id = 0;

    } else {

      host try_host;
      slice_manager::ret_code r;

      do {

	r = _slices->get_new_fid( pol, new_id, try_host );
	if( r != slice_manager::SM_SUCCESS &&
	    r != slice_manager::SM_TRYHOST ) {
	  return wfs_protocol::IOERR;
	}

	if( type != WFS_TYPE_DIRECTORY ) {
	  new_id |= WFS_FILE_MASK;
	} else {
	  new_id &= ~WFS_FILE_MASK;
	}

	if( type == WFS_TYPE_LINK ) {
	  new_id |= WFS_LINK_MASK;
	} else {
	  new_id &= ~WFS_LINK_MASK;
	}

#ifdef MAIL_HACK
      } while( r == slice_manager::SM_SUCCESS &&
	       type == WFS_TYPE_DIRECTORY &&
	       ((new_id & 0xf) == 0x1 || (new_id & 0xf) == 0x9) );
#else
    } while(0);
#endif

    if( r == slice_manager::SM_TRYHOST ) {
      new_id = 0; // host needs to create it
      file_node = make_node(try_host);
    } else {
      file_node = get_primary( new_id );
      if( file_node.id == 0 ) {
        return wfs_protocol::IOERR;
      }
      wfs_log(WFS_DBG_1, "Got new file id for %s: %016"PRIx64"\n",
          name.c_str(), new_id );
    }
  }

    new_fh = file_id_to_file_handle( new_id, dir_handle );

    if( file_node.id != dir_node.id ) {
      // TODO: make repeated_call?
      if( is_local(file_node) ) {
	s = _local_server->create_file( user, new_fh, type,
					permissions, final_fh );
	new_fh = final_fh;
      } else {
	rpcc *client = _clients->get_client( file_node, fhto(dir_handle)  );
	if( client == NULL ) {
	  s = rpc_const::timeout_failure;
	} else {
	  s = client->call( wfs_protocol::create_file,
			    user, new_fh, (uint16_t) type, permissions,
			    final_fh, fhto(dir_handle), "host");
	  _clients->done_with_client(file_node);
	  new_fh = final_fh;
	}
      }
      if( s != wfs_protocol::OK ) {
	continue;
      }
    }

    // link
    final_fh = new_fh;
    wfs_version new_version;
    s = wfs_client::link( new_fh, dir_handle, name,
			  (file_node.id == dir_node.id), false, type,
			  permissions, user, final_fh, new_version,
			  client );
    // repeat until we find an unused ID
  } while( s == wfs_protocol::ID_EXIST );

  new_fh = final_fh;

  return s;

}

wfs_protocol::status wfs_client::readlink( wfs_file_handle fh,
					   wfs_version version,
					   const std::string &user,
					   std::string &link ) {

  return readfile( fh, version, 0, WFS_BLOCK_SIZE, user, link );

}

wfs_protocol::status wfs_client::symlink( wfs_file_handle dir_handle,
					  std::string name,
					  std::string link,
					  const std::string &user,
					  wfs_file_handle &new_fh,
					  host *client ) {

  // create the file and write the link into it
  // make the permissions up
  static std::map<std::string, wfs_permission> link_perms;
  link_perms[ANYONE_IDENT] = WFS_PERMISSION_ALL;
  wfs_protocol::status s = create( dir_handle, name,
				   link_perms, WFS_TYPE_LINK,
				   user, new_fh, client );

  if( s != wfs_protocol::OK ) {
    return s;
  }

  wfs_version new_version;
  s = writefile( new_fh, wfs_version(), 0, link, user, new_version, client );

  if( s != wfs_protocol::OK ) {
    return s;
  }

  s = flush( new_fh, user, new_version, client );

  return s;

}

wfs_protocol::status wfs_client::link( wfs_file_handle fh,
				       wfs_file_handle dir_handle,
				       std::string name,
				       bool create_file,
				       bool link_file,
				       wfs_file_type type,
				       const std::map<std::string,
				       wfs_permission> &permissions,
				       const std::string &user,
				       wfs_file_handle &new_fh,
				       wfs_version &new_file_version,
				       host *client ) {

  // create the appropriate new directory entry
  // create the entry on the directory server
  wfs_node dir_node = get_primary( file_handle_to_file_id( dir_handle ) );
  wfs_node file_node = get_primary( file_handle_to_file_id( fh ) );
  if( dir_node.id == 0 ) {
    return wfs_protocol::IOERR;
  }

  host h;
  if( client != NULL ) {
    h = *client;
  }
  wfs_file_handle final_fh;

  wfs_protocol::status dir_status;

  wfs_protocol::create_file_struct cf;
  cf.create = create_file;
  cf.permissions = permissions;
  cf.type = type;
  // shouldn't ever need to create the file and link the file
  assert( !(create_file && link_file) );
  if( link_file && file_node.id == dir_node.id ) {
    cf.link = true;
  } else {
    cf.link = false;
  }
  wfs_protocol::file_and_version ret;

  wfs_log(WFS_DBG_1, "client::link %016"PRIx64" %s\n", new_fh.fid,
	    format_permissions(permissions).c_str() );

  if( is_local(dir_node) ) {
    dir_status = _local_server->create_dirent( user, dir_handle, name, fh, cf,
					       h, ret );
  } else {
    marshall m;
    m << user;
    m << dir_handle;
    m << name;
    m << fh;
    m << cf;
    m << h;
    dir_status = repeated_call( dir_handle, wfs_protocol::create_dirent,
				m, ret );
  }
  new_fh = ret.fh;
  new_file_version = ret.ver;

  if( dir_status == wfs_protocol::EXIST ) {
    wfs_log(WFS_DBG_1, "%s already exists in directory %016"PRIx64"\n", name.c_str(),
	      dir_handle.fid );
    // oops, a file with that name already exists, use the right handle
    return wfs_protocol::OK;
  } else if( dir_status != wfs_protocol::OK ) {
    return dir_status;
  }

  if( link_file && file_node.id != dir_node.id ) {
    wfs_protocol::status file_status;
    if( is_local(file_node) ) {
      file_status = _local_server->add_link( user, fh, 1, h, new_file_version );
    } else {
      marshall m;
      m << user;
      m << fh;
      m << 1;
      m << h;
      file_status = repeated_call( fh, wfs_protocol::add_link,
				   m, new_file_version );
    }
    return file_status;
  } else {
    return dir_status;
  }

}

wfs_protocol::status wfs_client::unlink( wfs_file_handle dir_handle,
					 std::string name, bool is_dir,
					 bool dec_link, const std::string &user,
					 wfs_protocol::file_and_version &fv,
					 host *client ) {

  // don't allow unlinks to happen in the .nodes directory
  if( is_meta_handle( dir_handle ) ) {
    return wfs_protocol::IOERR;
  }

  wfs_node dir_node = get_primary( file_handle_to_file_id( dir_handle ) );
  if( dir_node.id == 0 ) {
    return wfs_protocol::IOERR;
  }

  // if it's a directory, make sure it's empty first
  if( is_dir ) {
    wfs_file_handle fh;
    wfs_protocol::status ret = lookup( dir_handle, name, user, fh );
    if( ret != wfs_protocol::OK || isfile(fh) ) {
      return ret;
    }
    std::vector<wfs_storage::dirent> e;
    ret = readdir( fh, 0, 1, user, e );
    if( ret != wfs_protocol::OK ) {
      return ret;
    }
    if( e.size() > 0 ) {
      return wfs_protocol::NOTEMPTY;
    }
  }

  wfs_protocol::status ret;

  host h;
  if( client != NULL ) {
    h = *client;
  }

  if( is_local(dir_node) ) {
    ret = _local_server->unlink( user, dir_handle, name, is_dir, dec_link,
				 h, fv );
  } else {

    marshall m;
    m << user;
    m << dir_handle;
    m << name;
    m << is_dir;
    m << dec_link;
    m << h;
    ret = repeated_call( dir_handle, wfs_protocol::unlink, m, fv );

  }

  wfs_log(WFS_DBG_1, "Finished unlinking file %s from dir %016"PRIx64"\n", name.c_str(), dir_handle.fid );

  wfs_node file_node = get_primary( file_handle_to_file_id(fv.fh) );
  if( dec_link && dir_node.id != file_node.id ) {
    wfs_log(WFS_DBG_1, "unlink about to decrement num_links for file %016"PRIx64"\n",
	      fv.fh.fid );
    if( is_local(file_node) ) {
      ret = _local_server->add_link( user, fv.fh, -1, h, fv.ver );
    } else {
      marshall m;
      m << user;
      m << fv.fh;
      m << -1;
      m << h;
      ret = repeated_call( fv.fh, wfs_protocol::add_link, m, fv.ver );
    }
  }

  return ret;

  // TODO: reclaim storage space somehow on the file server

}

wfs_protocol::status wfs_client::readdir( wfs_file_handle fh,
					  unsigned long long ind,
					  unsigned int n,
					  const std::string &user,
					  std::vector<wfs_storage::dirent> &e,
					  host *client ) {

  if( is_meta_handle(fh) ) {

    std::vector<host> hosts;
    _slices->get_all_active_hosts(hosts);
    unsigned long long zind = 0;
    std::set<std::string> site_map;

    while( e.size() < n ) {
      wfs_storage::dirent entry;
      if( zind >= hosts.size() ) {
	break;
      }
      wfs_node next_node = make_node(hosts[zind]);
      if( fh.is_sites ) {
	if( site_map.find(next_node.site_id) == site_map.end() ) {
	  if( next_node.site_id != "" ) {
	    entry.name = next_node.site_id;
	    entry.inum = fh;
	    entry.inum.is_sites = false;
	    entry.inum.site = next_node.site_id;
	    site_map.insert(next_node.site_id);
	    e.push_back(entry);
	  }
	}
      } else if( fh.site == "" || next_node.site_id == fh.site ) {
	char name[100];
	sprintf( name, "%s:%d,%016"PRIx64"",
		 inet_ntoa( next_node.addr.sin_addr ),
		 htons(next_node.addr.sin_port), next_node.id );
	entry.name = std::string( name, strlen(name) );
	entry.inum = fh;
	entry.inum.is_nodes = false;
	entry.inum.site = "";
	entry.inum.node = next_node;
	e.push_back(entry);
      }
      zind++;
    }

    return wfs_protocol::OK;
  }

  wfs_node dir_node = get_primary( file_handle_to_file_id( fh ) );
  if( dir_node.id == 0 ) {
    return wfs_protocol::IOERR;
  }

  host h;
  if( client != NULL ) {
    h = *client;
  }

  wfs_protocol::status s;

  if( is_local(dir_node) ) {
    s = _local_server->readdir( user, fh, ind, n, h, e );
  } else {
    marshall m;
    m << user;
    m << fh;
    m << ind;
    m << n;
    m << h;
    s = repeated_call( fh, wfs_protocol::readdir, m, e );
  }

  // it's important that the node field propagates along with the handle
  // for directories
  for( unsigned i = 0; i < e.size(); i++ ) {
    e[i].inum.node = fh.node;
  }
  return s;

}

wfs_protocol::status wfs_client::rename( wfs_file_handle old_parent,
					 std::string old_name,
					 wfs_file_handle new_parent,
					 std::string new_name,
					 const std::string &user,
					 host *client ) {

  wfs_node old_node = get_primary( file_handle_to_file_id( old_parent ) );
  if( old_node.id == 0 ) {
    return wfs_protocol::IOERR;
  }
  wfs_node new_node = get_primary( file_handle_to_file_id( new_parent ) );
  if( new_node.id == 0 ) {
    return wfs_protocol::IOERR;
  }

  host h;
  if( client != NULL ) {
    h = *client;
  }

  if( old_node.id != new_node.id ) {

    // NOT GUARANTEED TO BE CONSISTENT
    // TODO: maybe optimize by sending these in parallel, and/or making
    // a version of unlink that returns the deleted name's file handle

    // find the file id
    wfs_file_handle old_fh;
    wfs_protocol::status s = lookup( old_parent, old_name, user, old_fh );

    if( s != wfs_protocol::OK ) {
      return s;
    }

    // unlink the old one
    wfs_protocol::file_and_version fv;
    s = unlink( old_parent, old_name, isdir(old_fh), false, user, fv, client );

    if( s != wfs_protocol::OK ) {
      return s;
    }

    wfs_protocol::create_file_struct cf;
    cf.create = false;
    //cf.permissions = 0;
    cf.link = false;

    // create the entry on the directory server
    if( is_local(new_node) ) {
      s = _local_server->create_dirent( user, new_parent, new_name, old_fh, cf,
					h, fv );
    } else {

      marshall m;
      m << user;
      m << new_parent;
      m << new_name;
      m << old_fh;
      m << cf;
      m << h;
      s = repeated_call( new_parent, wfs_protocol::create_dirent, m, fv );

    }
    return s;

  } else {

    // we can do the whole thing with a single RPC
    int dummy;
    int s;
    if( is_local(new_node) ) {
      s = _local_server->rename( user, old_parent, old_name,
				 new_parent, new_name,
				 h, dummy );
    } else {

      marshall m;
      m << user;
      m << old_parent;
      m << old_name;
      m << new_parent;
      m << new_name;
      m << h;
      s = repeated_call( new_parent, wfs_protocol::rename, m, dummy );

    }
    return s;

  }

}

wfs_protocol::status wfs_client::readfile_from( sockaddr_in addr,
						wfs_file_handle fh,
						wfs_version version,
						unsigned long long off,
						unsigned int n,
						const std::string &user,
						wfs_protocol::readinfo &info,
						bool fromclient) {

  if( fh.is_version ) {
    if( off < WFS_VERSION_FILE_SIZE ) {
      char ver[WFS_VERSION_FILE_SIZE];
      sprintf( ver, "%010d", version.version_num );
      info.buf = std::string( ver+off, WFS_VERSION_FILE_SIZE-off );
    }
    return wfs_protocol::OK;
  }

  rpcc::TO to = fhto(fh);
  to.use_tcp = true;
  // TODO: make repeated_call?
  wfs_protocol::status status;
  if( _local_server && addr.sin_port == _me.addr.sin_port &&
      addr.sin_addr.s_addr == _me.addr.sin_addr.s_addr ) {
    status = _local_server->readfile( user, fh, version, off, n, info );
  } else {
    wfs_node node = make_node( inet_ntoa(addr.sin_addr), htons(addr.sin_port) );
    rpcc *client = _clients->get_client( node, to );
    if( client == NULL ) {
      status = rpc_const::timeout_failure;
    } else {
      // This can be another client, so don't always require "host"
      status = client->call( wfs_protocol::readfile,
			     user,
			     fh, version,
			     off, n,
			     info, to, fromclient?NULL:"host");
      _clients->done_with_client(node);
    }
  }
  return status;
}

wfs_protocol::status wfs_client::readfile( wfs_file_handle fh,
					   wfs_version version,
					   unsigned long long off,
					   unsigned int n,
					   const std::string &user,
					   std::string &buf ) {

  wfs_node file_node = get_primary( file_handle_to_file_id( fh ) );
  if( file_node.id == 0 ) {
    return wfs_protocol::IOERR;
  }

  wfs_version v = version;
  if( fh.version.version_num >= 0 ) {
    v = fh.version;
  }

  wfs_protocol::status status;

  unsigned i = 1;
  bool repeat = false;
  int rep_level;
  std::vector<host> reps;
  slice_manager::ret_code r =
    _slices->get_slice_replicaset( wfs_slice::get_slice_id(fh.fid),
				   reps, rep_level );
  assert (r == slice_manager::SM_SUCCESS);

  wfs_protocol::readinfo info;

  // TODO: convert this into a repeated_call call, so I can reuse all that
  // code
  do {

    repeat = false;
    status = readfile_from( file_node.addr, fh, v, off, n, user, info );
    buf.swap(info.buf);  // avoid mem copies, hopefully!

    // don't need lax mode to do repeated lookups for file data
    if( status == -1 && i < (unsigned)rep_level && i < reps.size() ) {
      i++;
      file_node = make_node(reps[i-1]);
      repeat = true;
    } else if( status == wfs_protocol::STALE ) {
      // TODO: limit time or iterations
      repeat = true;
    }

  } while( repeat );

  return status;

}

#ifndef min
#define min(a, b) ((a) < (b) ? (a) : (b))
#endif

wfs_protocol::status wfs_client::writefile( wfs_file_handle fh,
					    wfs_version version,
					    unsigned long long off,
					    const std::string &buf,
					    const std::string &user,
					    wfs_version &new_version,
					    host *client ) {

  wfs_node file_node = get_primary(fh.fid);
  if( file_node.id == 0 ) {
    return wfs_protocol::IOERR;
  }

  int j = 0;
  int n = buf.length();

  host h;
  if( client != NULL ) {
    h = *client;
  }

  while( n > 0 ){
    int cc = min( n, wfs_protocol::MAXWRITE );
    std::string buf1( buf, j, cc );
    wfs_log(WFS_DBG_1, "calling write with offset %lld on file %016"PRIx64"\n", off+j,
	      fh.fid );

    wfs_protocol::status ret;

    if( is_local(file_node) ) {
      ret = _local_server->writefile( user, fh, (off+j), buf1, h, new_version );
    } else {
      marshall m;
      m << user;
      m << fh;
      m << (off+j);
      m << buf1;
      m << h;
      ret = repeated_call( fh, wfs_protocol::writefile, m, new_version );
    }

    if( ret != wfs_protocol::OK ) {
      // TODO: this could be a problem, because all subsequent writes
      // will just be forgotten
      return ret;
    }
    j += cc;
    n -= cc;
  }

  return wfs_protocol::OK;

}

wfs_protocol::status wfs_client::writefile_batch( wfs_file_handle fh,
						  std::vector<wfs_protocol::write> writes,
						  const std::string &user,
						  wfs_version &new_version,
						  host *client ) {

  wfs_node file_node = get_primary(fh.fid);
  if( file_node.id == 0 ) {
    return wfs_protocol::IOERR;
  }

  host h;
  if( client != NULL ) {
    h = *client;
  }

  // I don't think we need to care about sending too much data, now that
  // TCP is underneath the RPC layer

  wfs_protocol::status ret;

  if( is_local(file_node) ) {
    ret = _local_server->writefile_batch( user, fh, writes, h, new_version );
  } else {
    marshall m;
    m << user;
    m << fh;
    m << writes;
    m << h;
    ret = repeated_call( fh, wfs_protocol::writefile_batch, m, new_version );
  }

  return ret;

}

wfs_protocol::status wfs_client::flush( wfs_file_handle fh,
					const std::string &user,
					wfs_version &new_version,
					host *client ) {

  return wfs_protocol::OK;  // doesn't do any caching, no need to flush

}

wfs_protocol::status wfs_client::truncate( wfs_file_handle fh,
					   wfs_version version,
					   unsigned long long off,
					   const std::string &user,
					   wfs_version &new_version,
					   host *client ) {

  wfs_node file_node = get_primary( file_handle_to_file_id( fh ) );
  if( file_node.id == 0 ) {
    return wfs_protocol::IOERR;
  }

  host h;
  if( client != NULL ) {
    h = *client;
  }

  wfs_protocol::status ret;

  if( is_local(file_node) ) {
    ret = _local_server->truncate( user, fh, off, h, new_version );
  } else {
    marshall m;
    m << user;
    m << fh;
    m << off;
    m << h;
    ret = repeated_call( fh, wfs_protocol::truncate, m, new_version );
  }

  return ret;

}

wfs_protocol::status wfs_client::chmod( wfs_file_handle fh,
					const std::map<std::string,
					wfs_permission> &permissions,
					const std::string &user,
					wfs_version &new_version, host *client ) {

  wfs_node file_node = get_primary( file_handle_to_file_id( fh ) );
  if( file_node.id == 0 ) {
    return wfs_protocol::IOERR;
  }

  host h;
  if( client != NULL ) {
    h = *client;
  }

  wfs_protocol::status ret;

  if( is_local(file_node) ) {
    ret = _local_server->chmod( user, fh, permissions, h, new_version );
  } else {
    marshall m;
    m << user;
    m << fh;
    m << permissions;
    m << h;
    ret = repeated_call( fh, wfs_protocol::chmod, m, new_version );
  }

  return ret;

}
