/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang, 
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


// overdisk protocol

#ifndef __WFS_PROTOCOL_H__
#define __WFS_PROTOCOL_H__ 1

#include <arpa/inet.h>
#include <rpc.h>
#include <wfs_file.h>
#include <util/ident_util.h>
#include <wfs_storage.h>
#include <wfs_cacheloc.h>

class wfs_protocol {
 public:
  enum xxstatus { OK, RPCERR, NOENT, IOERR, EXIST, NOTDIR, ISDIR, NOTEMPTY, 
		  BADVERSION, STALE, OK_FLUSH, ID_EXIST, PERM, NOTMASTER };
  typedef int status;
  enum rpc_numbers {
    getfile = 0x5001 | rpc_const::VIVALDI_MASK,
    getdir = 0x5002 | rpc_const::VIVALDI_MASK,
    lookup = 0x5003 | rpc_const::VIVALDI_MASK,
    readdir = 0x5004,
    readfile = 0x5005,
    writefile = 0x5006,
    writefile_batch = 0x5007,
    create_dirent = 0x5008 | rpc_const::VIVALDI_MASK,
    create_file = 0x5009 | rpc_const::VIVALDI_MASK,
    truncate = 0x500a | rpc_const::VIVALDI_MASK,
    unlink = 0x500b | rpc_const::VIVALDI_MASK,
    addcacheloc = 0x500c | rpc_const::VIVALDI_MASK,
    getcacheloc = 0x500d | rpc_const::VIVALDI_MASK,
    invalidate = 0x500e,
    rename = 0x500f,
    chmod = 0x5010,
    renew_lease_batch = 0x5011 | rpc_const::VIVALDI_MASK,
    add_link = 0x5012 | rpc_const::VIVALDI_MASK,
    getfiles = 0x5013 | rpc_const::VIVALDI_MASK
  };

  // fit in a UDP packet
  static const int MAXREAD = 65536;
  static const int MAXWRITE = 65536;

  struct readinfo {
    std::string buf;
    std::vector<int32_t> block_map;
    std::vector<cacheloc_ent> others;
  };
  struct fileinfo {
    unsigned long long size;
    std::string perm_cues;
    wfs_version version;
    std::map<std::string,wfs_permission> permissions;
    unsigned long atime;
    unsigned long mtime;
    unsigned long ctime;
  };
  struct blockinfo {
    unsigned long long off;
    int size;
    wfs_version version;
  };
  struct dirinfo {
    unsigned long long size;
    std::string perm_cues;
    std::map<std::string,wfs_permission> permissions;
    unsigned long atime;
    unsigned long mtime;
    unsigned long ctime;
  };
  struct write {
    unsigned long long off;
    std::string data;
    uint16_t truncate;
  };
  struct create_file_struct {
    uint16_t create;
    uint16_t link;
    std::map<std::string,wfs_permission> permissions;
    uint16_t type;
  };
  struct lookupinfo {
    wfs_file_handle fh;
    std::vector<struct fileinfo> extra_info;
  };
  struct file_and_version {
    wfs_file_handle fh;
    wfs_version ver;
  };
};

inline unmarshall &
operator>>(unmarshall &u, wfs_protocol::file_and_version &e)
{
  u >> e.fh;
  u >> e.ver;
  return u;
}

inline marshall &
operator<<(marshall &m, wfs_protocol::file_and_version &e)
{
  m << e.fh;
  m << e.ver;
  return m;
}


inline unmarshall &
operator>>(unmarshall &u, wfs_protocol::readinfo &e)
{
  u >> e.buf;
  u >> e.block_map;
  u >> e.others;
  return u;
}

inline marshall &
operator<<(marshall &m, wfs_protocol::readinfo e)
{
  m << e.buf;
  m << e.block_map;
  m << e.others;
  return m;
}

inline unmarshall &
operator>>(unmarshall &u, wfs_storage::dirent &e)
{
  u >> e.name;
  u >> e.inum;
  return u;
}

inline marshall &
operator<<(marshall &m, wfs_storage::dirent e)
{
  m << e.name;
  m << e.inum;
  return m;
}

inline unmarshall &
operator>>(unmarshall &u, wfs_protocol::fileinfo &e)
{
  u >> e.size;
  u >> e.perm_cues;
  u >> e.version.version_num;
  u >> e.version.node_id;
  u >> e.permissions;
  u >> e.atime;
  u >> e.mtime;
  u >> e.ctime;
  return u;
}

inline marshall &
operator<<(marshall &m, wfs_protocol::fileinfo e)
{
  m << e.size;
  m << e.perm_cues;
  m << e.version.version_num;
  m << e.version.node_id;
  m << e.permissions;
  m << e.atime;
  m << e.mtime;
  m << e.ctime;
  return m;
}

inline unmarshall &
operator>>(unmarshall &u, std::vector<wfs_protocol::fileinfo> &e)
{
  int size;
  u >> size;
  for( int i = 0; i < size; i++ ) {
    wfs_protocol::fileinfo fin;
    u >> fin;
    e.push_back(fin);
  }
  return u;
}

inline marshall &
operator<<(marshall &m, std::vector<wfs_protocol::fileinfo> e)
{
  m << e.size();
  for( std::vector<wfs_protocol::fileinfo>::iterator i = e.begin(); 
       i != e.end(); i++ ) {
    m << *i;
  }
  return m;
}


inline unmarshall &
operator>>(unmarshall &u, wfs_protocol::blockinfo &e)
{
  u >> e.off;
  u >> e.size;
  u >> e.version.version_num;
  u >> e.version.node_id;
  return u;
}

inline marshall &
operator<<(marshall &m, wfs_protocol::blockinfo e)
{
  m << e.off;
  m << e.size;
  m << e.version.version_num;
  m << e.version.node_id;
  return m;
}
inline unmarshall &
operator>>(unmarshall &u, wfs_protocol::dirinfo &e)
{
  u >> e.size;
  u >> e.perm_cues;
  u >> e.permissions;
  u >> e.atime;
  u >> e.mtime;
  u >> e.ctime;
  return u;
}

inline marshall &
operator<<(marshall &m, wfs_protocol::dirinfo e)
{
  m << e.size;
  m << e.perm_cues;
  m << e.permissions;
  m << e.atime;
  m << e.mtime;
  m << e.ctime;
  return m;
}

inline unmarshall &
operator>>(unmarshall &u, wfs_protocol::write &e)
{
  u >> e.off;
  u >> e.data;
  u >> e.truncate;
  return u;
}

inline marshall &
operator<<(marshall &m, wfs_protocol::write e)
{
  m << e.off;
  m << e.data;
  m << e.truncate;
  return m;
}

inline marshall &
operator<<(marshall &m, cacheloc_ent s)
{
  m << s.s_addr;
  m << s.s_port;
  m << s.off;
  m << s.n;
  m << s.c;
  m << s.age;
#ifdef WFS_WITH_LIBSSH
  m << s.checksums;
#endif
  return m;
}

#ifdef WFS_WITH_LIBSSH
inline marshall &
operator<<(marshall &m, const cacheloc_ent::checksum &c) {
  for (unsigned lcv = 0; lcv < SHA256_DIGEST_LENGTH; lcv++) {
    m.rawbyte(c.data[lcv]);
  }
  return m;
}
#endif

inline unmarshall &
operator>>(unmarshall &u, cacheloc_ent &s)
{
  u >> s.s_addr;
  u >> s.s_port;
  u >> s.off;
  u >> s.n;
  u >> s.c;
  u >> s.age;
#ifdef WFS_WITH_LIBSSH
  u >> s.checksums;
#endif
  return u;
}

#ifdef WFS_WITH_LIBSSH
inline unmarshall &
operator>>(unmarshall &u, cacheloc_ent::checksum &c) {
  for (unsigned lcv = 0; lcv < SHA256_DIGEST_LENGTH; lcv++) {
    c.data[lcv] = u.rawbyte();
  }
  return u;
}
#endif

inline unmarshall &
operator>>(unmarshall &u, wfs_protocol::create_file_struct &e)
{
  u >> e.create;
  u >> e.link;
  u >> e.permissions;
  u >> e.type;
  return u;
}

inline marshall &
operator<<(marshall &m, wfs_protocol::create_file_struct e)
{
  m << e.create;
  m << e.link;
  m << e.permissions;
  m << e.type;
  return m;
}

inline unmarshall &
operator>>(unmarshall &u, wfs_protocol::lookupinfo &e)
{
  u >> e.fh;
  u >> e.extra_info;
  return u;
}

inline marshall &
operator<<(marshall &m, wfs_protocol::lookupinfo e)
{
  m << e.fh;
  m << e.extra_info;
  return m;
}


#endif // __WFS_PROTOCOL_H__
