/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang,
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


// overdisk server
// stores contents of each ID in ./ID

#include <host.h>
#include <fcntl.h>
#include <method_thread.h>
#include <sstream>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>

#include <wfs_metadata.h>
#include <wfs_misc.h>
#include <wfs_node.h>

#include "wfs_cent_cserver.h"
#include "wfs_server.h"

wfs_file_handle wfs_server::empty_fh;
wfs_file_handle wfs_maint::empty_fh;

wfs_server::wfs_server( std::string dir, wfs_perfile_locker *locker,
                        wfs_maint* maint, wfs_storage *storage,
			wfs_client_manager *clients, slice_manager *slices ) :
  _dir(dir), _locker(locker), _maint(maint), _storage(storage),
  _clients(clients), _slices(slices), _num_threads(0) {

  assert( pthread_mutex_init( &_is_lock, NULL ) == 0 );
  assert( pthread_mutex_init( &_nt_lock, NULL ) == 0 );
  assert( pthread_cond_init( &_nt_cond, NULL ) == 0 );

}

wfs_server::~wfs_server() {
  pthread_mutex_destroy( &_is_lock );
}

bool wfs_server::isfile( wfs_file_handle fh ) {
  return wfs_isfile(fh);
}

bool wfs_server::isdir( wfs_file_handle fh ) {
  return wfs_isdir(fh);
}

wfs_version wfs_server::read_version( wfs_file_handle fh ) {

  wfs_metadata *md = _storage->get_md(fh.fid);
  wfs_version v = md->version;
  _storage->release_md(md);
  return v;

}

/* xavid: If this file is versioned, copy out the current version before
   making changes.  There should be at most one
   save_old_version per (later) update_version (but not for a newly
   created file/directory).

   This assumes that fh is locked.
*/
const void wfs_server::save_old_version( wfs_file_handle fh ) {
  if (fh.get_cue( WFS_CUE_VERSIONING ).val == WFS_CUE_VERSIONED) {
    _storage->save_old_version(fh);
  }
}

const wfs_version wfs_server::update_version( wfs_file_handle fh) {
  wfs_metadata *md = _storage->get_md(fh.fid);
  wfs_version v = update_version( fh, md );
  _storage->set_and_release_md( fh.fid, md );
  return v;
}

// If the metadata was already loaded to update the version, don't reload it
const wfs_version wfs_server::update_version( wfs_file_handle fh,
					      wfs_metadata *md ) {

  wfs_node node = _maint->me();
  md->version.version_num++;
  md->version.node_id = node.id;
  return md->version;

}

int wfs_server::pingresponder( host h, host &dummy ) {
  return wcc_protocol::OK;
}

bool wfs_server::check_stale( wfs_file_handle fh ) {
  // wait for half the timeout, so the client won't needlessly fail
  int to_wait_ms = fhto(fh).to/2;
  if( to_wait_ms == -1 ) {
    to_wait_ms = 2500;
  }
  bool answer = _locker->is_stale( fh, to_wait_ms );
  // if it is stale, but the fh is lax, maybe we don't care.
  if( answer && fh.get_cue( WFS_CUE_CONSISTENCY ).val == WFS_CUE_EVENTUAL ) {
    answer = false;
  }
  return answer;
}

inline void set_thread_id( wfs_file_handle fh ) {
  if( wfs_dbg_block(2) ) {
    wfs_set_thread_id( fh.get_cue(WFS_CUE_ID).val );
  }
}

inline void clear_thread_id() {
  if( wfs_dbg_block(2) ) {
    wfs_set_thread_id(0);
  }
}

bool wfs_server::can_handle( wfs_file_handle fh ) {

  // if I am the master, handle it
  // if I am not the master, and it's eventual, and I'm in the replica set,
  //   handle it
  // otherwise do not handle it
  if( _slices->is_primary_for_file(fh.fid) ) {
    return true;
  }

  int32_t consistency = fh.get_cue( WFS_CUE_CONSISTENCY ).val;
  if( consistency == WFS_CUE_EVENTUAL ) {
    std::vector<host> reps;
    int rep_num;
    _slices->get_slice_replicaset( wfs_slice::get_slice_id(fh.fid),
				   reps, rep_num );
    // am i on the list?
    for( int i = 0; i < rep_num && i < (int)reps.size(); i++ ) {
      if( _maint->me().id == make_node(reps[i]).id ) {
	return true;
      }
    }
  }

  wfs_log(WFS_DBG_1, "Not master for %016"PRIx64"! Returning.\n", fh.fid );

  return false;

}

int wfs_server::getfile( const std::string &user,
			 wfs_file_handle fh,
			 host client,
                         wfs_protocol::fileinfo &fin ) {

  set_thread_id(fh);

  wfs_log(WFS_DBG_1, "getfile %016"PRIx64" %s\n", fh.fid, fh.print().c_str() );

  // We assume that if we have the file handle, we can stat the file,
  // so no permission check

  if( !can_handle(fh) ) {
    return wfs_protocol::NOTMASTER;
  }

  if( check_stale(fh) ) {
    wfs_log(WFS_DBG_OFF, "%016"PRIx64" is stale!\n", fh.fid );
    return wfs_protocol::STALE;
  }

  struct stat sb;
  if( !_storage->stat_object( fh, &sb ) ) {
    return wfs_protocol::NOENT;
  }
  // TODO: figure out why this cast is needed.  Without it, fin.size is all
  // wrong.
  fin.size = (int) sb.st_size;
  fin.atime = sb.st_atime;
  fin.mtime = sb.st_mtime;
  fin.ctime = sb.st_ctime;

  _locker->lock_file(fh);

  wfs_metadata *md = _storage->get_md(fh.fid);
  fin.perm_cues = md->perm_cues;
  fin.version = md->version;
  fin.permissions = md->permissions;
  _storage->release_md(md);
  wfs_log(WFS_DBG_2, "permissions are %s\n",
	    format_permissions(fin.permissions).c_str());

  // grant a lease
  // TODO: make sure this host was contacted during the last
  // invalidation; if not, must return OK_FLUSH
  if( client.addr != 0 ) {
    wfs_node node = make_node( inet_ntoa(client.sin().sin_addr),
                               ntohs(client.port) );
    wfs_log(WFS_DBG_1, "Granting lease on %016"PRIx64" to %s:%d\n", fh.fid,
              inet_ntoa(client.sin().sin_addr),
              ntohs(client.port) );
    _leases.add_lease( fh, node, fhlease(fh), fhstaletime(fh) );
  }

  _locker->unlock_file(fh);

  clear_thread_id();
  return wfs_protocol::OK;

}

//bulk request for file info
int wfs_server::getfiles( const std::string &user,
			 std::vector<wfs_file_handle> fhs,
			 host client,
                         std::vector<wfs_protocol::fileinfo> &fins ) {

  for( std::vector<wfs_file_handle>::iterator i = fhs.begin(); i != fhs.end(); i++) {
    wfs_file_handle fh = *i;
    set_thread_id(fh);

    wfs_log(WFS_DBG_1, "getfile %016"PRIx64" %s\n", fh.fid, fh.print().c_str() );

    // We assume that if we have the file handle, we can stat the file,
    // so no permission check

    if( !can_handle(fh) ) {
      return wfs_protocol::NOTMASTER;
    }

    if( check_stale(fh) ) {
      wfs_log(WFS_DBG_OFF, "%016"PRIx64" is stale!\n", fh.fid );
      return wfs_protocol::STALE;
    }

    struct stat sb;
    if( !_storage->stat_object( fh, &sb ) ) {
      return wfs_protocol::NOENT;
    }

    wfs_protocol::fileinfo fin;
    // TODO: figure out why this cast is needed.  Without it, fin.size is all
    // wrong.
    fin.size = (int) sb.st_size;
    fin.atime = sb.st_atime;
    fin.mtime = sb.st_mtime;
    fin.ctime = sb.st_ctime;

    _locker->lock_file(fh);

    wfs_metadata *md = _storage->get_md(fh.fid);
    fin.perm_cues = md->perm_cues;
    fin.version = md->version;
    fin.permissions = md->permissions;
    _storage->release_md(md);
    wfs_log(WFS_DBG_2, "permissions are %s\n",
	      format_permissions(fin.permissions).c_str());

    // grant a lease
    // TODO: make sure this host was contacted during the last
    // invalidation; if not, must return OK_FLUSH
    if( client.addr != 0 ) {
      wfs_node node = make_node( inet_ntoa(client.sin().sin_addr),
				 ntohs(client.port) );
      wfs_log(WFS_DBG_1, "Granting lease on %016"PRIx64" to %s:%d\n", fh.fid,
          inet_ntoa(client.sin().sin_addr),
          ntohs(client.port) );
      _leases.add_lease( fh, node, fhlease(fh), fhstaletime(fh) );
    }

    fins.push_back(fin);
    _locker->unlock_file(fh);

    clear_thread_id();
  }
  return wfs_protocol::OK;

}


int wfs_server::getdir( const std::string &user,
			wfs_file_handle fh,
                        host client,
                        wfs_protocol::dirinfo &fin ) {

  set_thread_id(fh);
  wfs_log(WFS_DBG_1, "getdir %016"PRIx64"\n", fh.fid );

  if( !can_handle(fh) ) {
    return wfs_protocol::NOTMASTER;
  }

  // We assume that if we have the file handle, we can stat the dir,
  // so no permission check

  if( check_stale(fh) ) {
    wfs_log(WFS_DBG_OFF, "%016"PRIx64" is stale!\n", fh.fid );
    return wfs_protocol::STALE;
  }

  struct stat sb;

  if( !_storage->stat_object( fh, &sb ) ) {

#ifdef MAIL_HACK
    /**********************
     * Hack for OSDI!!!!!!!!!!
     * Create this directory on the fly
     ***********************/

    _locker->lock_file(fh);
    fh.set_cue( WFS_CUE_DIR_LOC_POLICY, WFS_CUE_KEEP_TOGETHER );

    std::map<std::string,wfs_permission> world_readable;
    world_readable[ANYONE_IDENT] = WFS_PERMISSION_ALL;

    icreate_file( fh, world_readable );
    assert( _storage->object_exists(fh) );

    wfs_file_handle new_fh = fh;

    // last digit of 1 == maildir
    if( (fh.fid & 0xF) == 0x1 ) {
      new_fh.fid -= 8;
      while( _storage->object_exists(new_fh) ) {
	new_fh.fid -= 16; // must always end in 0xd
      }
      icreate_file( new_fh, world_readable );
      icreate_dirent( fh, "Maildir", new_fh, client, empty_fh );
    } else if( (fh.fid & 0xF) == 0x9 ) {

      do {
	new_id( new_fh, WFS_TYPE_DIRECTORY, fh, true );
      } while( (new_fh.fid & 0xf) == 0x1 || (new_fh.fid & 0xf) == 0x9 );
      icreate_file( new_fh, world_readable );
      icreate_dirent( fh, "new", new_fh, client, empty_fh );
      do {
	new_id( new_fh, WFS_TYPE_DIRECTORY, fh, true );
      } while( (new_fh.fid & 0xf) == 0x1 || (new_fh.fid & 0xf) == 0x9 );
      icreate_file( new_fh, world_readable );
      icreate_dirent( fh, "cur", new_fh, client, empty_fh );
      do {
	new_id( new_fh, WFS_TYPE_DIRECTORY, fh, true );
      } while( (new_fh.fid & 0xf) == 0x1 || (new_fh.fid & 0xf) == 0x9 );
      icreate_file( new_fh, world_readable );
      icreate_dirent( fh, "tmp", new_fh, client, empty_fh );

      // temporaries
      // don't replicate, that's lame
      new_fh.set_cue( WFS_CUE_REP_LEVEL, 1 );
      do {
	new_id( new_fh, WFS_TYPE_DIRECTORY, fh, true );
      } while( (new_fh.fid & 0xf) == 0x1 || (new_fh.fid & 0xf) == 0x9 );
      icreate_file( new_fh, world_readable );
      icreate_dirent( fh, "index", new_fh, client, empty_fh );
    }

    _locker->unlock_file(fh);

#else
      clear_thread_id();
      return wfs_protocol::NOENT;
#endif

  } else {

#ifdef MAIL_HACK
    if( ((int) sb.st_size) == 0 && (fh.fid & 0xF) == 0x9 ) {

      std::map<std::string,wfs_permission> world_readable;
      world_readable[ANYONE_IDENT] = WFS_PERMISSION_ALL;

      _locker->lock_file(fh);
      fh.set_cue( WFS_CUE_DIR_LOC_POLICY, WFS_CUE_KEEP_TOGETHER );
      wfs_file_handle new_fh = fh;
      do {
	new_id( new_fh, WFS_TYPE_DIRECTORY, fh, true );
      } while( (new_fh.fid & 0xf) == 0x1 || (new_fh.fid & 0xf) == 0x9 );
      icreate_file( new_fh, world_readable );
      icreate_dirent( fh, "new", new_fh, client, empty_fh );
      do {
	new_id( new_fh, WFS_TYPE_DIRECTORY, fh, true );
      } while( (new_fh.fid & 0xf) == 0x1 || (new_fh.fid & 0xf) == 0x9 );
      icreate_file( new_fh, world_readable );
      icreate_dirent( fh, "cur", new_fh, client, empty_fh );
      do {
	new_id( new_fh, WFS_TYPE_DIRECTORY, fh, true );
      } while( (new_fh.fid & 0xf) == 0x1 || (new_fh.fid & 0xf) == 0x9 );
      icreate_file( new_fh, world_readable );
      icreate_dirent( fh, "tmp", new_fh, client, empty_fh );

      // temporaries
      // don't replicate, that's lame
      new_fh.set_cue( WFS_CUE_REP_LEVEL, 1 );
      do {
	new_id( new_fh, WFS_TYPE_DIRECTORY, fh, true );
      } while( (new_fh.fid & 0xf) == 0x1 || (new_fh.fid & 0xf) == 0x9 );
      icreate_file( new_fh, world_readable );
      icreate_dirent( fh, "index", new_fh, client, empty_fh );

      _locker->unlock_file(fh);
    }
#endif

  }

  _locker->lock_file(fh);

  wfs_metadata *md = _storage->get_md(fh.fid);
  fin.perm_cues = md->perm_cues;
  fin.permissions = md->permissions;
  fin.atime = sb.st_atime;
  fin.mtime = sb.st_mtime;
  fin.ctime = sb.st_ctime;
  fin.size = (int) sb.st_size;
  _storage->release_md(md);

  // grant a lease
  // TODO: make sure this host was contacted during the last
  // invalidation; if not, must return OK_FLUSH
  if( client.addr != 0 ) {
    wfs_node node = make_node( inet_ntoa(client.sin().sin_addr),
                               ntohs(client.port) );
    wfs_log(WFS_DBG_1, "Granting lease on %016"PRIx64" to %s:%d\n", fh.fid,
        inet_ntoa(client.sin().sin_addr),
        ntohs(client.port) );
    _leases.add_lease( fh, node, fhlease(fh), fhstaletime(fh) );
  }

  _locker->unlock_file(fh);
  clear_thread_id();
  return wfs_protocol::OK;

}

// internal version of lookup
// Assume the directory is locked already

wfs_file_handle wfs_server::ilookup( wfs_file_handle dh, std::string name ) {

  wfs_file_handle fh = dh; // copies the cues
  fh.inc_depth();

  // this should get done on the client, just making sure
  // (could be setting debug cue)
  if( search_and_set_cues( name, fh ) ) {
    if (fh.fid != 0) {
      wfs_log(WFS_DBG_1, "fid is 0 after successful sasc!\n");
    }
    return fh;
  }

  wfs_file_handle empty;
  wfs_file_id fid = 0;
  if( !_storage->lookup( dh, name, fid ) ) {
    wfs_log(WFS_DBG_1, "lookup failed due to storage failure!\n");
    return empty;
  } else {
    if (fid != 0) {
      wfs_log(WFS_DBG_1, "fid is not 0 after storage lookup!\n");
    }
    fh.fid = fid;
    return fh;
  }

}

int wfs_server::lookup( const std::string &user,
			wfs_file_handle dh, std::string name,
			host client, wfs_protocol::lookupinfo &li ) {

  set_thread_id(dh);
  wfs_log(WFS_DBG_1, "lookup %016"PRIx64" %s\n", dh.fid, name.c_str());

  if( !can_handle(dh) ) {
    return wfs_protocol::NOTMASTER;
  }

  if( check_stale(dh) ) {
    wfs_log(WFS_DBG_OFF, "%016"PRIx64" is stale!\n", dh.fid );
    clear_thread_id();
    return wfs_protocol::STALE;
  }

  _locker->lock_file(dh);

  // We assume dh exists, and since we only check positions on the
  // directory, whether name exists or not doesn't matter for permissions
  wfs_metadata *md = _storage->get_md(dh.fid);
  if (!has_permission(md->permissions, user, WFS_PERMISSION_READ)) {
    _locker->unlock_file(dh);
    _storage->release_md(md);
    clear_thread_id();
    return wfs_protocol::PERM;
  }
  _storage->release_md(md);

  li.fh = ilookup(dh, name);
  _locker->unlock_file(dh);

  // grant a lease
  // TODO: make sure this host was contacted during the last
  // invalidation; if not, must return OK_FLUSH
  if( client.addr != 0 ) {
    wfs_node node = make_node( inet_ntoa(client.sin().sin_addr),
                               ntohs(client.port) );
    wfs_log(WFS_DBG_1, "Granting lease on (%s) to %s:%d\n", dh.print().c_str(),
              inet_ntoa(client.sin().sin_addr),
              ntohs(client.port) );
    _leases.add_lease( dh, node, fhlease(dh), fhstaletime(dh) );
  }

  if( li.fh.fid == 0 ) {
    clear_thread_id();
    return wfs_protocol::NOENT;
  }

  wfs_log(WFS_DBG_1, "Lookup succeeded: %s\n", li.fh.print().c_str() );

  // now, if we happen to also be the primary for this file, go ahead and
  // return the info for it
  if( isfile(li.fh) ) {
    wfs_protocol::fileinfo finfo;
    if( getfile( user, li.fh, client, finfo ) == wfs_protocol::OK ) {
      li.extra_info.push_back(finfo);
    }
  } else {
    wfs_protocol::dirinfo dinfo;
    if( getdir( user, li.fh, client, dinfo ) == wfs_protocol::OK ) {
      wfs_protocol::fileinfo finfo;
      finfo.perm_cues = dinfo.perm_cues;
      finfo.permissions = dinfo.permissions;
      finfo.size = dinfo.size;
      finfo.atime = dinfo.atime;
      finfo.ctime = dinfo.ctime;
      finfo.mtime = dinfo.mtime;
      finfo.version = wfs_version();
      li.extra_info.push_back(finfo);
    }
  }

  clear_thread_id();
  return wfs_protocol::OK;

}

void wfs_server::call_invalidate( wfs_file_handle fh,
                                  wfs_lease_manager::leaser leaser ) {

  wfs_node node = leaser.node;
  time_t time_left_sec = leaser.time_left;
  int32_t stale_time = leaser.stale_time;

  wfs_log(WFS_DBG_1, "Calling invalidate for %016"PRIx64" on node %s:%d, to=%d, st=%d\n",
            fh.fid, inet_ntoa(node.addr.sin_addr), ntohs(node.addr.sin_port),
            (int)time_left_sec, stale_time );

  pthread_mutex_lock( &_is_lock );

  bool sleeping = _invalidate_sleepers[node.id][fh.fid];
  pthread_mutex_unlock( &_is_lock );

  // we can safely return, since another thread will invalidate later
  if( sleeping ) {
    wfs_log(WFS_DBG_2, "invalidate: sleeping, so returning\n");
    pthread_mutex_lock( &_nt_lock );
    _num_threads--;
    pthread_cond_signal( &_nt_cond );
    pthread_mutex_unlock( &_nt_lock );
    pthread_exit( NULL );
    return;
  }

  if( stale_time > 0 ) {
    pthread_mutex_lock( &_is_lock );
    _invalidate_sleepers[node.id][fh.fid] = true;
    pthread_mutex_unlock( &_is_lock );
    wfs_log(WFS_DBG_1, "About to sleep for %d before invalidating lease on %016"PRIx64" for %016"PRIx64"\n",
              stale_time, fh.fid, node.id );
    usleep(stale_time*1000);
    pthread_mutex_lock( &_is_lock );
    _invalidate_sleepers[node.id][fh.fid] = false;
    pthread_mutex_unlock( &_is_lock );
  }

  void *ret;
  rpcc *client = _clients->get_client( node, rpcc::to(time_left_sec*1000)  );
  if( client == NULL ) {
    wfs_log(WFS_DBG_1, "invalidate: client is NULL\n");
    ret = (void *) rpc_const::timeout_failure;
  } else {
    wfs_log(WFS_DBG_2, "invalidate: calling invalidate on client\n");
    int dummy;
    ret = (void *) client->call( wfs_protocol::invalidate,
				 fh, dummy, rpcc::to(time_left_sec*1000) );
    _clients->done_with_client(node);
  }

  wfs_log(WFS_DBG_1, "Invalidate done for fh %s\n", fh.print().c_str() );

  pthread_mutex_lock( &_nt_lock );
  _num_threads--;
  pthread_cond_signal( &_nt_cond );
  pthread_mutex_unlock( &_nt_lock );

  pthread_exit( ret );

}

void wfs_server::invalidate_leases( wfs_file_handle fh, host modifier,
                                    wfs_file_handle other_fh ) {
  std::vector<wfs_lease_manager::leaser> leasers;
  _leases.leasing_nodes( fh, leasers );

  wfs_node node_mod;
  if( modifier.addr != 0 ) {
    node_mod = make_node( inet_ntoa(modifier.sin().sin_addr),
                          ntohs(modifier.port) );
    _leases.invalidate_lease( fh, &node_mod );
  } else {
    _leases.invalidate_lease( fh );
  }

  std::vector<pthread_t> pthreads;

  // release the lock over the RPCs
  _locker->unlock_file(fh);
  if( other_fh.fid != 0 && other_fh.fid != fh.fid ) {
    _locker->unlock_file(other_fh);
  }

  // TODO: if the server crashed and restarted, it will technically need to
  // wait the maximum lease time before returning from this function,
  // because it may have lost important lease state.  Not sure if this is
  // worth doing.

  for( unsigned i = 0; i < leasers.size(); i++ ) {

    if( modifier.addr != 0 && leasers[i].node.id == node_mod.id ) {
      wfs_log(WFS_DBG_1, "Not invalidating modifying node %016"PRIx64"\n", node_mod.id );
      continue; // don't invalidate the modifier
    } else {
      wfs_log(WFS_DBG_1, "Invalidating non-modifying node %016"PRIx64" (not mod %s:%d,%016"PRIx64")\n",
	      leasers[i].node.id, inet_ntoa(modifier.sin().sin_addr),
	      ntohs(modifier.port), node_mod.id );
    }

    // if this thread already has something sleeping, skip now
    pthread_mutex_lock( &_is_lock );
    bool sleeping = _invalidate_sleepers[leasers[i].node.id][fh.fid];
    pthread_mutex_unlock( &_is_lock );
    // we can safely move on, since another thread will invalidate later
    if( sleeping ) {
      wfs_log(WFS_DBG_1, "Invalidation: sleeping!\n" );
      continue;
    }

    bool join = (leasers[i].stale_time <= 0);

    wfs_log(WFS_DBG_1, "About to invalidate, join=%d, staletime=%d\n", join,
	      leasers[i].stale_time );

    pthread_mutex_lock( &_nt_lock );
    while( _num_threads >= 50 ) {
      pthread_cond_wait( &_nt_cond, &_nt_lock );
    }
    _num_threads++;
    pthread_mutex_unlock( &_nt_lock );

    pthread_t pth = method_thread( this, !join,
                                   &wfs_server::call_invalidate, fh,
                                   leasers[i] );

    if( join && pth != 0 ) {
      pthreads.push_back(pth);
    }

  }

  // wait for the threads to finish
  for( unsigned i = 0; i < pthreads.size(); i++ ) {

    wfs_maint_protocol::status s;
    void *tmptr;
    pthread_join( pthreads[i], &tmptr );
    s = (int)(long)tmptr;
    if( s != wfs_maint_protocol::OK ) {
      // TODO: mark somewhere that this lease was never invalidated
      wfs_log(WFS_DBG_1, "invalidate_leases: lease was never invalidated (returned %d!)\n",
	       s);
    }

  }

  wfs_log(WFS_DBG_1, "Done with the invalidates\n" );

  // take it back
  if( other_fh.fid == fh.fid || other_fh.fid == 0 ) {
    _locker->lock_file(fh);
  } else if( other_fh.fid < fh.fid ) {
    assert( other_fh.fid != 0 );
    _locker->lock_file(other_fh);
    _locker->lock_file(fh);
  } else {
    assert( other_fh.fid != 0 );
    _locker->lock_file(fh);
    _locker->lock_file(other_fh);
  }

}

int wfs_server::invalidate_and_sync( wfs_file_handle fh, host modifier,
                                     wfs_file_handle other_fh,
				     std::string append_data ) {

  // invalidate any held leases before continuing
  invalidate_leases( fh, modifier, other_fh );

  if( fhsynclevel(fh) <= 1 ) {
    _maint->immediate_replica_enqueue( fh );
  } else {
    int s = _maint->immediate_replica_sync( fh, other_fh, append_data );
    if( s != 0 ) {
      wfs_log(WFS_DBG_1, "invalidate_and_sync: error %d!\n",s);
      return wfs_protocol::IOERR;
    }
  }

  return wfs_protocol::OK;

}

int wfs_server::icreate_dirent( wfs_file_handle dh, std::string name,
                                wfs_file_handle new_fh,
				host client,
                                wfs_file_handle other_dh ) {

  // assume dir is locked and name is correct and exists

  wfs_log(WFS_DBG_1, "icreate_dirent %016"PRIx64" %s %016"PRIx64"\n", dh.fid, name.c_str(), new_fh.fid);

  std::string append_data = "";

  if( dh.get_cue(WFS_CUE_CONSISTENCY).val == WFS_CUE_EVENTUAL &&
      !_storage->object_exists(dh) ) {

    // in lax mode, go ahead and create the file if it doesn't yet exist
    std::map<std::string,wfs_permission> world_readable;
    world_readable[ANYONE_IDENT] = WFS_PERMISSION_ALL;
    icreate_file( dh, world_readable );

  }

  if( !_storage->link( dh, name, new_fh, append_data ) ) {
    wfs_log(WFS_DBG_1, "icreate_dirent: error!\n");
    return wfs_protocol::IOERR;
  }

  update_version(dh);

  return invalidate_and_sync( dh, client, other_dh, append_data );

}

int wfs_server::create_dirent( const std::string &user,
			       wfs_file_handle dh, std::string name,
                               wfs_file_handle new_fh,
			       wfs_protocol::create_file_struct cf,
			       host client,
                               wfs_protocol::file_and_version &final_file ) {

  set_thread_id(dh);

  wfs_log(WFS_DBG_1, "create_dirent %016"PRIx64" %s %016"PRIx64"\n", dh.fid, name.c_str(), new_fh.fid);

  if( !can_handle(dh) ) {
    return wfs_protocol::NOTMASTER;
  }

  if( check_stale(dh) ) {
    wfs_log(WFS_DBG_OFF, "%016"PRIx64" is stale!\n", dh.fid );
    clear_thread_id();
    return wfs_protocol::STALE;
  }

  // make the file first
  if( cf.create ) {

    if( new_fh.fid == 0 ) {
      new_id( new_fh, cf.type, dh, true );
    }

    wfs_log(WFS_DBG_1, "create_dirent creating empty file first: %016"PRIx64"\n", new_fh.fid );
    _locker->lock_file(new_fh);
    // TODO: do we care about the status returns of these?  It's
    // just an empty file
    wfs_protocol::status s = icreate_file( new_fh, cf.permissions );
    if( s != wfs_protocol::OK ) {
      _locker->unlock_file(new_fh);
      clear_thread_id();
      return s;
    }

    if( fhsynclevel(new_fh) <= 1 ) {
      _maint->immediate_replica_enqueue( new_fh );
    } else {
      _maint->immediate_replica_sync( new_fh );
    }
    _locker->unlock_file(new_fh);
  } else if( cf.link ) {

    if( !can_handle(new_fh) ) {
      clear_thread_id();
      return wfs_protocol::NOTMASTER;
    }

    // inc the link
    _locker->lock_file(new_fh);

    if( !_storage->object_exists(new_fh) ) {
      _locker->unlock_file(new_fh);
      clear_thread_id();
      return wfs_protocol::NOENT;
    }

    wfs_metadata *md = _storage->get_md(new_fh.fid);

    // Save an old version if this is versioned
    save_old_version(new_fh);
    md->num_links++;
    final_file.ver = update_version( new_fh, md );
    _storage->set_and_release_md( new_fh.fid, md );
    invalidate_and_sync( new_fh, client );

    _locker->unlock_file(new_fh);

  }

  _locker->lock_file(dh);

  // -! new_fh might end up created even if the permissions check fails!
  wfs_metadata *md = _storage->get_md(dh.fid);
  if (!has_permission(md->permissions, user, WFS_PERMISSION_WRITE)) {
    _locker->unlock_file(dh);
    _storage->release_md(md);
    clear_thread_id();
    return wfs_protocol::PERM;
  }
  _storage->release_md(md);

  wfs_file_handle xfh = ilookup( dh, name );
  if( xfh.fid != 0 ) {
    // the new file should win -- delete the old guy
    if( isfile(new_fh) && isfile(xfh) ) {
      // TODO: the old guy will not have his num_links decremented sadly ...
      wfs_protocol::status s = iunlink( dh, name, client, empty_fh,
					false /*don't sync!*/ );
      if( s != wfs_protocol::OK ) {
        _locker->unlock_file(dh);
	clear_thread_id();
        return s;
      }
    } else {
      // directories are trickier since there could be important meta-data
      // already in the existing dir that the client is expecting to see
      _locker->unlock_file(dh);
      final_file.fh = xfh;
      clear_thread_id();
      return wfs_protocol::EXIST;
    }
  }

  int s = icreate_dirent( dh, name, new_fh, client );

  _locker->unlock_file(dh);

  final_file.fh = new_fh;

  wfs_log(WFS_DBG_1, "create done %016"PRIx64" %s %016"PRIx64"\n", dh.fid, name.c_str(), new_fh.fid);

  clear_thread_id();
  return s;

}

int wfs_server::iunlink( wfs_file_handle dh, std::string name, host client,
                         wfs_file_handle other_dh, bool sync ) {

  // assume dir is locked and name is correct and exists

  wfs_log(WFS_DBG_1, "iunlink %016"PRIx64" %s\n", dh.fid, name.c_str() );

  // Save an old version if this directory is versioned
  save_old_version(dh);

  if( !_storage->object_exists(dh) ) {
    return wfs_protocol::NOENT;
  } else if( !_storage->unlink( dh, name ) ) {
    wfs_log(WFS_DBG_1, "iunlink: error!\n");
    return wfs_protocol::IOERR;
  }

  update_version(dh);

  // Sometimes we don't want to sync because we still need to make
  // other updates atomically to the directory, and we want to make
  // sure the directory stays locked!
  if( sync ) {
    return invalidate_and_sync( dh, client, other_dh );
  } else {
    return wfs_protocol::OK;
  }

}

int wfs_server::unlink( const std::string &user, wfs_file_handle dh,
			std::string name, int is_dir, int dec_link,
                        host client, wfs_protocol::file_and_version &fv ) {

  set_thread_id(dh);

  if( !can_handle(dh) ) {
    return wfs_protocol::NOTMASTER;
  }

  wfs_log(WFS_DBG_1, "remove %016"PRIx64" %s\n", dh.fid, name.c_str() );

  if( check_stale(dh) ) {
    wfs_log(WFS_DBG_OFF, "%016"PRIx64" is stale!\n", dh.fid );
    clear_thread_id();
    return wfs_protocol::STALE;
  }

  _locker->lock_file(dh);

  wfs_metadata *md = _storage->get_md(dh.fid);
  if (!has_permission(md->permissions, user, WFS_PERMISSION_WRITE)) {
    _locker->unlock_file(dh);
    _storage->release_md(md);
    clear_thread_id();
    return wfs_protocol::PERM;
  }
  _storage->release_md(md);

  fv.fh = ilookup(dh, name);
  if( fv.fh.fid == 0 ) {
    _locker->unlock_file(dh);
    clear_thread_id();
    wfs_log(WFS_DBG_2, "unlink failed due to lookup failure!\n");
    return wfs_protocol::NOENT;
  }

  // if we're the master for this file as while, decrement the link
  if( dec_link && can_handle(fv.fh) ) {

    _locker->lock_file(fv.fh);

    wfs_log(WFS_DBG_1, "unlink is decrementing the link directly for file %016"PRIx64"\n",
	      fv.fh.fid );

    if( !_storage->object_exists(fv.fh) ) {
      _locker->unlock_file(fv.fh);
      clear_thread_id();
      return wfs_protocol::NOENT;
    }

    wfs_metadata *md = _storage->get_md(fv.fh.fid);

    // Save an old version if this is versioned
    save_old_version(fv.fh);
    md->num_links--;
    int num_links = md->num_links;
    fv.ver = update_version( fv.fh, md );
    _storage->set_and_release_md( fv.fh.fid, md );
    invalidate_and_sync( fv.fh, client );

    if( num_links == 0 ) {
      _storage->delete_object(fv.fh);
    }

    _locker->unlock_file(fv.fh);

  }

  bool fh_is_dir = isdir(fv.fh);
  if( fh_is_dir != is_dir ) {
    _locker->unlock_file(dh);
    clear_thread_id();
    if( is_dir ) {
      return wfs_protocol::NOTDIR;
    } else {
      return wfs_protocol::ISDIR;
    }
  }

  int s = iunlink( dh, name, client );

  _locker->unlock_file(dh);

  wfs_log(WFS_DBG_1, "remove complete %016"PRIx64" %s\n", dh.fid, name.c_str() );

  clear_thread_id();
  return s;

}

int wfs_server::icreate_file( wfs_file_handle new_fh,
			      const std::map<std::string,
			      wfs_permission> &permissions ) {

  if( !_storage->create_new_object( new_fh ) ) {
    return wfs_protocol::ID_EXIST;
  }

  wfs_log(WFS_DBG_1, "icreate_file %016"PRIx64" %s\n", new_fh.fid,
	    format_permissions(permissions).c_str() );

  std::string cues = new_fh.perm_cues_string();
  wfs_metadata *md = _storage->get_md(new_fh.fid);
  md->perm_cues = cues;
  md->version.version_num++;
  md->version.node_id = _maint->me().id;
  md->permissions = permissions;
  // every file must start out with 1 link, right?
  md->num_links = 1;
  _storage->set_and_release_md( new_fh.fid, md );

  return wfs_protocol::OK;

}

void wfs_server::new_id( wfs_file_handle &new_fh, uint16_t type,
			 wfs_file_handle parent, bool force_local ) {

    rep_policy pol( fhreplevel(new_fh), rep_policy::STRICT,
		    _maint->me().site_id, fhrepsites(new_fh) );

    if( parent.fid != 0 &&
	parent.get_cue(WFS_CUE_DIR_LOC_POLICY).val ==
	WFS_CUE_KEEP_TOGETHER ) {
      _slices->get_policy_for_slice( wfs_slice::get_slice_id(parent.fid),
				     pol );
      // use the desired replevel and repsite; just use the parent
      // policy for the primary site
      // TODO: this is kind of a bug, since the policy could then
      // have a different replica set.  but can't keep the replica set
      // the same without having it in the replication policy...
      pol.rep_num = fhreplevel(new_fh);
      pol.site_num = fhrepsites(new_fh);
    }

    host tryhost;

    do {
      slice_manager::ret_code r =
	_slices->get_new_fid( pol, new_fh.fid, tryhost, force_local );
      assert( r >= 0 );

      if( type != WFS_TYPE_DIRECTORY ) {
	new_fh.fid |= WFS_FILE_MASK;
      } else {
	new_fh.fid &= ~WFS_FILE_MASK;
      }

      if( type == WFS_TYPE_LINK ) {
	new_fh.fid |= WFS_LINK_MASK;
      } else {
	new_fh.fid &= ~WFS_LINK_MASK;
      }

#ifdef MAIL_HACK
    } while( (type == WFS_TYPE_DIRECTORY &&
	      ((new_fh.fid & 0xf) == 0x1 || (new_fh.fid & 0xf) == 0x9)) ||
	     _storage->object_exists(new_fh) );
#else
    } while( _storage->object_exists(new_fh) );
#endif

}

int wfs_server::create_file( const std::string &user,
			     wfs_file_handle new_fh, uint16_t type,
			     std::map<std::string,
			     wfs_permission> permissions,
			     wfs_file_handle &final_fh ) {

  set_thread_id(new_fh);

  if( new_fh.fid == 0 ) {
    new_id( new_fh, type, new_fh, true );
  }

  if( !can_handle(new_fh) ) {
    return wfs_protocol::NOTMASTER;
  }

  final_fh = new_fh;

  _locker->lock_file(new_fh);

  // Creating a file can be done by anyone

  wfs_protocol::status s = icreate_file( new_fh, permissions );

  if( s != wfs_protocol::OK ) {
    _locker->unlock_file(new_fh);
    clear_thread_id();
    return s;
  }

  // TODO: can we avoid this sync somehow?  It's just an empty file,
  // this is kind of lame
  if( fhsynclevel(new_fh) <= 1 ) {
    _maint->immediate_replica_enqueue( new_fh );
  } else {
    int s = _maint->immediate_replica_sync( new_fh );
    if( s != 0 ) {
      _locker->unlock_file(new_fh);
      clear_thread_id();
      wfs_log(WFS_DBG_1, "create_file: error!\n");
      return wfs_protocol::IOERR;
    }
  }

  _locker->unlock_file(new_fh);
  clear_thread_id();
  return wfs_protocol::OK;
}

int wfs_server::readdir( const std::string &user,
			 wfs_file_handle fh, unsigned long long ind,
                         unsigned int n,
			 host client,
                         std::vector<wfs_storage::dirent> &e ) {

  set_thread_id(fh);
  wfs_log(WFS_DBG_1, "readdir %016"PRIx64" ind %lld\n", fh.fid, ind );

  if( !can_handle(fh) ) {
    return wfs_protocol::NOTMASTER;
  }

  if( check_stale(fh) ) {
    wfs_log(WFS_DBG_OFF, "%016"PRIx64" is stale!\n", fh.fid );
    clear_thread_id();
    return wfs_protocol::STALE;
  }

  _locker->lock_file(fh);

  if( !_storage->readdir( fh, ind, n, e ) ) {
    _locker->unlock_file(fh);
    clear_thread_id();
    return wfs_protocol::NOENT;
  }

  // Check permissions after checking existance
  wfs_metadata *md = _storage->get_md(fh.fid);
  if (!has_permission(md->permissions, user, WFS_PERMISSION_READ)) {
    _locker->unlock_file(fh);
    _storage->release_md(md);
    clear_thread_id();
    return wfs_protocol::PERM;
  }
  _storage->release_md(md);

  _locker->unlock_file(fh);

  // grant a lease
  // TODO: make sure this host was contacted during the last
  // invalidation; if not, must return OK_FLUSH
  if( client.addr != 0 ) {
    wfs_node node = make_node( inet_ntoa(client.sin().sin_addr),
                               ntohs(client.port) );
    wfs_log(WFS_DBG_1, "Granting lease on %016"PRIx64" to %s:%d\n", fh.fid,
              inet_ntoa(client.sin().sin_addr),
              ntohs(client.port) );
    _leases.add_lease( fh, node, fhlease(fh), fhstaletime(fh) );
  }

  clear_thread_id();
  return wfs_protocol::OK;

}

int wfs_server::rename(  const std::string &user,
			 wfs_file_handle old_parent, std::string old_name,
                        wfs_file_handle new_parent, std::string new_name,
			host client, int &dummy ) {

  set_thread_id(old_parent);
  wfs_log(WFS_DBG_1, "rename %016"PRIx64" %s %016"PRIx64" %s\n", old_parent.fid, old_name.c_str(),
            new_parent.fid, new_name.c_str() );

  if( !can_handle(old_parent) || !can_handle(new_parent) ) {
    return wfs_protocol::NOTMASTER;
  }

  if( check_stale(old_parent) || check_stale(new_parent) ) {
    wfs_log(WFS_DBG_OFF, "%016"PRIx64" or %016"PRIx64" is stale!\n", old_parent.fid, new_parent.fid );
    clear_thread_id();
    return wfs_protocol::STALE;
  }

  // get locks on these directories in the correct order

  if( old_parent.fid == new_parent.fid ) {
    _locker->lock_file( old_parent );
  } else {
    if( old_parent.fid < new_parent.fid ) {
      _locker->lock_file( old_parent );
      _locker->lock_file( new_parent );
    } else {
      _locker->lock_file( new_parent );
      _locker->lock_file( old_parent );
    }
  }

  wfs_metadata *omd = _storage->get_md(old_parent.fid);
  wfs_metadata *nmd = _storage->get_md(new_parent.fid);
  if (!has_permission(omd->permissions, user, WFS_PERMISSION_WRITE)
      || !has_permission(nmd->permissions, user, WFS_PERMISSION_WRITE)) {
    _locker->unlock_file(old_parent);
    if( old_parent.fid != new_parent.fid ) {
      _locker->unlock_file( new_parent );
    }
    _storage->release_md(omd);
    _storage->release_md(nmd);
    clear_thread_id();
    return wfs_protocol::PERM;
  }
  _storage->release_md(omd);
  _storage->release_md(nmd);

  wfs_file_handle old_fh = ilookup( old_parent, old_name );

  if( old_fh.fid == 0 ) {
    _locker->unlock_file( old_parent );
    if( old_parent.fid != new_parent.fid ) {
      _locker->unlock_file( new_parent );
    }
    clear_thread_id();
    return wfs_protocol::NOENT;
  }

  int s = iunlink( old_parent, old_name, client, new_parent,
		   false /* don't sync!*/ );

  if( s != wfs_protocol::OK ) {
    _locker->unlock_file( old_parent );
    if( old_parent.fid != new_parent.fid ) {
      _locker->unlock_file( new_parent );
    }
    clear_thread_id();
    return s;
  }

  // it's ok to overwrite a file that already exists, just go ahead and
  // delete it first
  wfs_file_handle new_fh = ilookup( new_parent, new_name );
  if( new_fh.fid != 0 ) {
    s = iunlink( new_parent, new_name, client, old_parent,
		 false /*don't sync!*/ );
    if( s != wfs_protocol::OK ) {
      _locker->unlock_file( old_parent );
      if( old_parent.fid != new_parent.fid ) {
        _locker->unlock_file( new_parent );
      }
      clear_thread_id();
      return s;
    }
  }

  s = icreate_dirent( new_parent, new_name, old_fh, client, old_parent );
  // now sync the old_parent, since we didn't before
  s = invalidate_and_sync( old_parent, client, new_parent );

  _locker->unlock_file( old_parent );
  if( old_parent.fid != new_parent.fid ) {
    _locker->unlock_file( new_parent );
  }
  clear_thread_id();
  return s;

}

int wfs_server::readfile( const std::string &user,
			  wfs_file_handle fh, wfs_version version,
                          unsigned long long off, unsigned int n,
			  wfs_protocol::readinfo &info ) {

  set_thread_id(fh);
  wfs_log(WFS_DBG_1, "wfs_server::readfile %016"PRIx64" ver %d node %016"PRIx64" off=%lld n=%d uname=%s\n",
	    fh.fid, version.version_num, version.node_id, off, n ,
	    user.c_str());

  if( !can_handle(fh) ) {
    return wfs_protocol::NOTMASTER;
  }

  if( check_stale(fh) ) {
    wfs_log(WFS_DBG_OFF, "%016"PRIx64" is stale!\n", fh.fid );
    clear_thread_id();
    return wfs_protocol::STALE;
  }

  _locker->lock_file(fh);

  wfs_version disk_ver = read_version(fh);
  wfs_version use_version;
  bool possible_bad_version = false;
  if( disk_ver != version ) {
    // +xavid: If the version doesn't match, try to fetch an old version
    use_version = version;
    wfs_log(WFS_DBG_1, "using version %d and node %016"PRIx64" instead of disk version %d and node %016"PRIx64"\n",
	      version.version_num, version.node_id,
	      disk_ver.version_num, disk_ver.node_id );
    possible_bad_version = true;
  }

  if( !_storage->object_exists( fh, use_version.version_num ) ) {
    _locker->unlock_file(fh);
    clear_thread_id();
    if( possible_bad_version ) {
      return wfs_protocol::BADVERSION;
    } else {
      return wfs_protocol::NOENT;
    }
  }

  // Only check permissions if file exists
  wfs_metadata *md = _storage->get_md(fh.fid, use_version.version_num);
  if (!has_permission(md->permissions, user, WFS_PERMISSION_READ)) {
    _locker->unlock_file(fh);
    _storage->release_md(md);
    clear_thread_id();
    return wfs_protocol::PERM;
  }
  _storage->release_md(md);

  bool read = _storage->readfile( fh, use_version, off, n, info.buf );

  if(read) {
    wfs_log(WFS_DBG_1, "wfs_server::readfile finished %016"PRIx64" ver %d off=%lld n=%d\n",
        fh.fid, version.version_num, off, n );
  }

  _locker->unlock_file(fh);
  clear_thread_id();

  if( !read ) {
    wfs_log(WFS_DBG_1,"read_file: error!\n");
    return wfs_protocol::IOERR;
  } else {
    return wfs_protocol::OK;
  }

}

int wfs_server::do_write( wfs_file_handle fh,
                          std::vector<unsigned long long> off,
                          std::vector<std::string> buf,
			  const std::string &user,
			  wfs_version &new_version ) {

  wfs_log(WFS_DBG_1, "do_write %016"PRIx64" numwrites=%u uname=%s\n", fh.fid,
	    (unsigned)off.size(), user.c_str());

  if( !can_handle(fh) ) {
    return wfs_protocol::NOTMASTER;
  }

  if( check_stale(fh) ) {
    wfs_log(WFS_DBG_OFF, "%016"PRIx64" is stale!\n", fh.fid );
    return wfs_protocol::STALE;
  }

  // Save an old version if this is versioned
  save_old_version(fh);

  if( !_storage->object_exists(fh) ) {
    return wfs_protocol::NOENT;
  }

  // Check perm after checking existance
  wfs_metadata *md = _storage->get_md(fh.fid);
  if (!has_permission(md->permissions, user, WFS_PERMISSION_WRITE)) {
    _storage->release_md(md);
    return wfs_protocol::PERM;
  }

  if( !_storage->writefile( fh, off, buf ) ) {
    wfs_log(WFS_DBG_1, "do_write: error!\n");
    return wfs_protocol::IOERR;
  }

  new_version = update_version(fh, md);
  _storage->set_and_release_md( fh.fid, md );
  return wfs_protocol::OK;

}

int wfs_server::writefile( const std::string &user,
			   wfs_file_handle fh, unsigned long long off,
                           const std::string buf,
			   host client, wfs_version &new_version ) {

  set_thread_id(fh);
  _locker->lock_file(fh);

  std::vector<unsigned long long> offv;
  offv.push_back(off);
  std::vector<std::string> bufv;
  bufv.push_back(buf);
  int s = do_write( fh, offv, bufv, user, new_version );

  if( s == wfs_protocol::OK ) {

    s = invalidate_and_sync( fh, client );
    _locker->unlock_file(fh);
    clear_thread_id();
    return s;

  } else {
    _locker->unlock_file(fh);
    clear_thread_id();
    //return wfs_protocol::IOERR;
    return s;
  }

}

int wfs_server::do_truncate( wfs_file_handle fh, unsigned long long off,
			     const std::string &user,
                             wfs_version &new_version ) {

  set_thread_id(fh);
  wfs_log(WFS_DBG_1, "truncate called\n" );

  if( !can_handle(fh) ) {
    return wfs_protocol::NOTMASTER;
  }

  if( check_stale(fh) ) {
    wfs_log(WFS_DBG_OFF, "%016"PRIx64" is stale!\n", fh.fid );
    clear_thread_id();
    return wfs_protocol::STALE;
  }

  if( !_storage->object_exists(fh) ) {
    return wfs_protocol::NOENT;
  }

  // Check perm after checking existance
  wfs_metadata *md = _storage->get_md(fh.fid);
  if (!has_permission(md->permissions, user, WFS_PERMISSION_WRITE)) {
    _storage->release_md(md);
    return wfs_protocol::PERM;
  }

  // Save an old version if this is versioned
  save_old_version(fh);

  if( _storage->truncate( fh, off ) ) {
    new_version = update_version(fh, md);
    _storage->set_and_release_md( fh.fid, md );
    clear_thread_id();
    return wfs_protocol::OK;
  } else {
    _storage->release_md(md);
    clear_thread_id();
    wfs_log(WFS_DBG_1, "do_truncate: error!\n");
    return wfs_protocol::IOERR;
  }

}

int wfs_server::truncate( const std::string &user,
			  wfs_file_handle fh, unsigned long long off,
			  host client, wfs_version &new_version ) {

  set_thread_id(fh);
  _locker->lock_file(fh);
  int s = do_truncate( fh, off, user, new_version );

  if( s == wfs_protocol::OK ) {

    s = invalidate_and_sync( fh, client );
    _locker->unlock_file(fh);
    clear_thread_id();
    return s;

  } else {
    _locker->unlock_file(fh);
    clear_thread_id();
    return s;
  }

}

int wfs_server::chmod( const std::string &user,
		       wfs_file_handle fh,
		       std::map<std::string,wfs_permission> permissions,
		       host client,
                       wfs_version &new_version ) {

  set_thread_id(fh);
  wfs_log(WFS_DBG_1, "chmod called %016"PRIx64" %s\n", fh.fid,
	   format_permissions(permissions).c_str() );

  if( !can_handle(fh) ) {
    return wfs_protocol::NOTMASTER;
  }

  if( check_stale(fh) ) {
    wfs_log(WFS_DBG_OFF, "%016"PRIx64" is stale!\n", fh.fid );
    clear_thread_id();
    return wfs_protocol::STALE;
  }

  _locker->lock_file(fh);

  if( !_storage->object_exists(fh) ) {
    _locker->unlock_file(fh);
    clear_thread_id();
    return wfs_protocol::NOENT;
  }

  // Check perm after checking existance
  wfs_metadata *md = _storage->get_md(fh.fid);
  if (!has_permission(md->permissions, user, WFS_PERMISSION_ADMIN)) {
    _locker->unlock_file(fh);
    wfs_log(WFS_DBG_1, "user %s does not have permission to chmod!",
	     user.c_str());
    _storage->release_md(md);
    clear_thread_id();
    return wfs_protocol::PERM;
  }

  // Save an old version if this is versioned
  save_old_version(fh);

  // Change mentioned permissions, leave unmentioned ones alone
  std::map<std::string,wfs_permission>::const_iterator i;
  for (i = permissions.begin(); i != permissions.end(); i++)
    md->permissions[i->first] = i->second;

  wfs_log(WFS_DBG_1, "chmod: resulting permissions were %s\n",
	   format_permissions(md->permissions).c_str() );

  new_version = update_version( fh, md );
  _storage->set_and_release_md( fh.fid, md );
  int s = invalidate_and_sync( fh, client );
  _locker->unlock_file(fh);
  clear_thread_id();
  return s;
}

int wfs_server::add_link( const std::string &user,
			  wfs_file_handle fh, int num_add,
			  host client, wfs_version &new_version ) {

  set_thread_id(fh);
  wfs_log(WFS_DBG_1, "add_link called %016"PRIx64" %d\n", fh.fid, num_add );

  if( !can_handle(fh) ) {
    return wfs_protocol::NOTMASTER;
  }

  if( check_stale(fh) ) {
    wfs_log(WFS_DBG_OFF, "%016"PRIx64" is stale!\n", fh.fid );
    clear_thread_id();
    return wfs_protocol::STALE;
  }

  _locker->lock_file(fh);

  if( !_storage->object_exists(fh) ) {
    _locker->unlock_file(fh);
    clear_thread_id();
    return wfs_protocol::NOENT;
  }

  wfs_metadata *md = _storage->get_md(fh.fid);

  // Save an old version if this is versioned
  save_old_version(fh);
  md->num_links += num_add;
  int num_links = md->num_links;
  new_version = update_version( fh, md );
  _storage->set_and_release_md( fh.fid, md );
  int s = invalidate_and_sync( fh, client );

  if (s != wfs_protocol::OK) {
    wfs_log(WFS_DBG_1, "add_link.invalidate_and_sync got a %d!\n",s);
  }

  // if it's down to 0 links, delete
  if( num_links == 0 ) {
    _storage->delete_object(fh);
  }

  _locker->unlock_file(fh);
  clear_thread_id();
  return s;
}

int wfs_server::writefile_batch( const std::string &user,
				 wfs_file_handle fh,
                                 std::vector<wfs_protocol::write> writes,
				 host client, wfs_version &new_version ) {

  set_thread_id(fh);
  wfs_log(WFS_DBG_1, "Starting batched write for file %016"PRIx64" and size %zu\n",
            fh.fid, writes.size() );

  int ret = wfs_protocol::OK;
  bool any_success = false;

  _locker->lock_file(fh);

  for( unsigned i = 0; i < writes.size(); i++ ) {
    wfs_protocol::write w = writes[i];
    if( w.truncate ) {
      ret = do_truncate( fh, w.off, user, new_version );
    } else {
      std::vector<unsigned long long> offv;
      std::vector<std::string> bufv;
      do {
        offv.push_back(writes[i].off);
        bufv.push_back(writes[i].data);
      } while( (i+1 < writes.size()) && !writes[i+1].truncate && ++i );
      // -? This means that existance and permissions are checked n times,
      //    when they only need to be checked once... maybe
      //    do_write/do_truncate could be refactored?
      ret = do_write( fh, offv, bufv, user, new_version );
    }

    if( ret != wfs_protocol::OK ) {
      break;
    } else {
      any_success = true;
    }

  }

  if( any_success || ret == wfs_protocol::OK ) {
    ret = invalidate_and_sync( fh, client );
  }

  _locker->unlock_file(fh);

  wfs_log(WFS_DBG_1, "Returning from write batch, s=%d\n", ret );

  clear_thread_id();
  return ret;

}

int wfs_server::getcacheloc( wfs_file_handle fh, wfs_protocol::blockinfo b,
                             Coord c, int list_sz,
                             std::vector<cacheloc_ent> &list ) {

  set_thread_id(fh);
  _cache_loc.get_close_nodes( file_handle_to_file_id( fh ), b.version,
                              NULL, c, b.off, (int) b.size, list, list_sz);
  wfs_log(WFS_DBG_1, "getcacheloc called on fh.fid %016"PRIx64" v %d done list %zu/%d off %llu n %d\n",
           file_handle_to_file_id( fh ), b.version.version_num,
           list.size(), list_sz, b.off, b.size);

#ifdef WFS_WITH_LIBSSH
  if (wfs_dbg_block(WFS_DBG_2)) {
    wfs_log(WFS_DBG_2, "Returning %d fish:\n",list.size());
    std::vector<cacheloc_ent>::iterator it;
    for (it = list.begin(); it != list.end(); it++) {
      wfs_log(WFS_DBG_2, "\tMumble: ");
      std::map<unsigned,cacheloc_ent::checksum>::iterator cit;
      for (cit = it->checksums.begin(); cit != it->checksums.end(); cit++) {
        wfs_log(WFS_DBG_2,"\t\t#%u ",cit->first);
      }
      unsigned blocks = 0;
      for (int lcv = 0; lcv < it->n.size(); lcv++) {
        blocks += it->n[lcv]/WFS_BLOCK_SIZE;
      }
      if (blocks != it->checksums.size()) {
        wfs_log(WFS_DBG_1,"\t\t(! %u < %u !)",it->checksums.size(),blocks);
      }
    }
  }
#endif
  clear_thread_id();
  return wfs_protocol::OK;
}

int wfs_server::addcacheloc( wfs_file_handle fh, wfs_version version,
                             cacheloc_ent loc, int &dummy ) {

  set_thread_id(fh);
  if( wfs_dbg_block(WFS_DBG_1) ) {
    in_addr sin_addr;
    sin_addr.s_addr = loc.s_addr;
    wfs_log(WFS_DBG_1, "addcacheloc called on %016"PRIx64", addr %s:%d off %llu n %llu c %s\n",
            file_handle_to_file_id(fh),
            inet_ntoa(sin_addr), ntohs(loc.s_port), loc.off[0],
            loc.n[0], loc.c.print_coord().c_str() );
  }

#ifdef WFS_WITH_LIBSSH
  // Add checksum if we're secure (checksum isn't useful without secure chans)
  if (rpc_is_secure()) {
    // TODO: We should really reuse checksums if we're already caching
    //       this block, but I don't understand the system well enough
    //       to implement that right now
    loc.checksums.clear();
    for (unsigned lcv = 0; lcv < loc.off.size(); lcv++) {
      assert ( loc.off[lcv] % WFS_BLOCK_SIZE == 0);
      assert ( loc.n[lcv] % WFS_BLOCK_SIZE == 0);
      for (unsigned block = 0; block < loc.n[lcv] / WFS_BLOCK_SIZE; block++) {
	wfs_protocol::readinfo info;
	int res = readfile(ROOT_IDENT, fh, version,
			   loc.off[lcv]+block*WFS_BLOCK_SIZE,
			   WFS_BLOCK_SIZE,
			   info);
	if (res != wfs_protocol::OK) {
	  wfs_log(WFS_DBG_1, "Got an error %d while checksumming block %u after %llu!\n",
		   res,block,loc.off[lcv]);
	  return res;
	}
	cacheloc_ent::checksum cs;
	SHA256((unsigned char *)info.buf.c_str(), info.buf.length(),
	    cs.data);
	unsigned gblock = loc.off[lcv]/WFS_BLOCK_SIZE+block;
	loc.checksums[gblock] = cs;
	wfs_log(WFS_DBG_1, "btchecksum for gblock #%u starts with %llx; data starts with %llx and is len %u\n",
	    gblock,*((unsigned long long *)cs.data),
	    *((unsigned long long *)info.buf.c_str()),
	    info.buf.size());
	/*char fn[42];
	  snprintf(fn,42,"/tmp/%llx", *((unsigned long long *)cs.data));
	  int fd = open(fn,O_WRONLY | O_CREAT | O_TRUNC, 0644);
	  write(fd,info.buf.c_str(),info.buf.size());
	  close(fd);*/
      }
    }
  }
#endif

  std::vector<cacheloc_ent> list;
  list.push_back(loc);
  _cache_loc.update_cacheloc( file_handle_to_file_id( fh ), version, list);
  clear_thread_id();
  return wfs_protocol::OK;
}

int wfs_server::renew_lease_batch( std::vector<wfs_file_handle> fhs,
				   host client, int &dummy ) {

  wfs_log(WFS_DBG_1, "Renewing %zu leases for node %s:%d\n", fhs.size(),
	    inet_ntoa(client.sin().sin_addr),
	    ntohs(client.port) );

  wfs_node node = make_node( inet_ntoa(client.sin().sin_addr),
			     ntohs(client.port) );
  _leases.renew_leases( node, fhs );
  dummy = 0;
  return wfs_protocol::OK;
}
