#!/usr/bin/perl -w
#
# Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang, 
#                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
#     		     Massachusetts Institute of Technology,
# 		     New York University
# 
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
# 
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


# simple tests of create, lookup, readdir, unlink, read, write.
# just on a single client.

use strict;

my $verbose = 0;

if($#ARGV != 0){
    print STDERR "Usage: fs-test-1.pl directory\n";
    exit(1);
}
my $dir = $ARGV[0];

my $seq = 0;

my $files = { };
my @dead;

for(my $iters = 0; $iters < 200; $iters++){
    createone();
}

for(my $iters = 0; $iters < 300; $iters++){
    if(rand() < 0.1){
        livecheck();
    }
    if(rand() < 0.1){
        deadcheck();
    }
    if(rand() < 0.02){
        dircheck();
    }
    if(rand() < 0.5){
        createone();
    }
    if(rand() < 0.5){
        createagain();
    }
    if(rand() < 0.1){
        moveone();
    }
    if(rand() < 0.4){
        deleteone();
    }
}

dircheck();
while(deleteone()){
}
dircheck();
printf "fs-test-1.pl: passed all tests!\n";
exit(0);

sub randstr {
    my($len) = @_;
    my $x = "";
    for(my $i = 0; $i < $len; $i++){
	$x .= sprintf("%c", ord('a') + int(rand(26)));
    }
    return $x;
}

sub createone {
    my $name = "file-" . randstr(40) . "-$$-" . $seq;
    $seq = $seq + 1;
    my $contents = randstr(int(rand(20)));
    print "create $name\n" if $verbose;
    if(!open(F, ">$dir/$name")){
        print STDERR "fs-test-1: cannot create $dir/$name : $!\n";
        exit(1);
    }
    print F $contents;
    close(F);
    $files->{$name} = $contents;
}

sub createagain {
    my @a = keys(%$files);
    return if $#a < 0;
    my $i = int(rand($#a + 1));
    my $k = $a[$i];
    print "re-create $k\n" if $verbose;
    if(!open(F, ">$dir/$k")){
        print STDERR "fs-test-1: cannot re-create $dir/$k : $!\n";
        exit(1);
    }     
    my $contents = randstr(int(rand(20)));
    print F $contents;
    close(F);
    $files->{$k} = $contents;
}

sub moveone {
    my @a = keys(%$files);
    return if $#a < 0;
    my $name = "file-" . randstr(40) . "-$$-" . $seq;
    $seq = $seq + 1;
    my $i = int(rand($#a + 1));
    my $k = $a[$i];
    print "move $k to $name\n" if $verbose;
    if(!rename( "$dir/$k", "$dir/$name" )){
	print STDERR "fs-test-1: rename failed, $k to $name : $!\n";
	exit(1);
    }
    $files->{$name} = $files->{$k};
    delete $files->{$k};
    push(@dead, $k);
}

# make sure all the live files are there,
# and that all the dead files are not there.
sub dircheck {
    print "dircheck\n" if $verbose;
    opendir(D, $dir);
    my %h;
    my $f;
    while(defined($f = readdir(D))){
        if(!defined($h{$f})){
            $h{$f} = 0;
        }
        $h{$f} = $h{$f} + 1;
    }
    closedir(D);

    foreach $f (keys(%$files)){
        if(!defined($h{$f})){
            print STDERR "fs-test-1.pl: $f is not in the directory listing\n";
            exit(1);
        }
        if($h{$f} > 1){
            print STDERR "fs-test-1.pl: $f appears more than once in the directory\n";
            exit(1);
        }
    }

    foreach $f (@dead){
        if(defined($h{$f})){
            print STDERR "fs-test-1.pl: $f is dead but in directory listing\n";
            exit(1);
        }
    }
}

sub livecheck {
    my @a = keys(%$files);
    return if $#a < 0;
    my $i = int(rand($#a + 1));
    my $k = $a[$i];
    print "livecheck $k\n" if $verbose;
    if(!open(F, "$dir/$k")){
        print STDERR "fs-test-1: cannot open $dir/$k : $!\n";
        exit(1);
    }
    my $x = "";
    sysread(F, $x, 100000);
    if($x ne $files->{$k}){
	print STDERR "fs-test-1: $dir/$k has wrong contents\n";
        printf( STDERR "wanted %d: $files->{$k}\n", length($files->{$k}));
        printf( STDERR "got %d: $x\n", length($x));
        exit(1);
    }
        
    close(F);
    if( ! -f "$dir/$k" ){
	print STDERR "fs-test-1: $dir/$k is not of type file\n";
	exit(1);
    }
    if(open(F, ">$dir/$k/xx")){
	print STDERR "fs-test-1: $dir/$k acts like a directory, not a file\n";
        exit(1);
    }
}

sub deadcheck {
    my $name = "file-$$-" . $seq;
    $seq = $seq + 1;
    print "check-not-there $name\n" if $verbose;
    if(open(F, "$dir/$name")){
        print STDERR "fs-test-1: $dir/$name exists but should not\n";
        exit(1);
    }

    return if $#dead < 0;
    my $i = int(rand($#dead + 1));
    my $k = $dead[$i];
    return if defined($files->{$k}); # ???
    print "deadcheck $k\n" if $verbose;
    if(rand(1.0) < 0.5){
        if(open(F, $dir . "/" . $k)){
            print STDERR "fs-test-1: $dir/$k exists but should not\n";
            exit(1);
        }
    } else {
        if(unlink($dir . "/" . $k)){
            print STDERR "fs-test-1: $dir/$k unlinkable but should not exist\n";
            exit(1);
        }
    }
}

sub deleteone {
    my @a = keys(%$files);
    return 0 if $#a < 0;
    my $i = int(rand($#a + 1));
    my $k = $a[$i];
    print "delete $k\n" if $verbose;
    if(unlink($dir . "/" . $k) == 0){
        print STDERR "fs-test-1: unlink $k failed: $!\n";
        exit(1);
    }
    delete $files->{$k};
    push(@dead, $k);
    return 1;
}
