/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang, 
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/**
 * Stress test secure RPCs.  Create a server that receives big packets,
 * and fork a bunch of child processes to make calls to the servers.
 * Kill the clients at random times
 */

#include <arpa/inet.h>
#include <assert.h>
#include <rpc.h>
#include <wfs_misc.h>
#include <wfs_log.h>

// server-side handlers. they must be methods of some class
// to simplify rpcs::reg(). a server process can have handlers
// from multiple classes.
class srv {
public:
  int handle_big( const int a, std::string b, std::string &r );
};

int srv::handle_big( const int len, std::string in, std::string &r ) {
  r = std::string( (size_t)len, 'x' );
  return 0;
}

size_t random_size() {
  return (size_t) random() % (15*1024);
}

void client_attack( rpcc *client, int num ) {
  for( int i = 0; i < num; i++ ) {
    size_t in_len = random_size();
    std::string s( in_len, 'y' );
    std::string r;
    int len = random_size();
    printf( "%d: Input length %d, output length %d\n", getpid(), in_len, len );
    int status = client->call( 25, len, s, r );
    printf( "%d: status %d, output length %d\n", getpid(), status, r.size() );
    assert( status == 0 && r.size() == (unsigned) len ); 
  }
}

struct sockaddr_in dst;

void * do_client(void *x) {

  pid_t pid = fork();

  if( pid == 0 ) {
    rpcc client(dst);
    rpcc::TO to;
    to.to = 10000;
    assert( client.bind(to) == 0 );
    client_attack( &client, 1000 );
    exit(0);
  } else {
    int usecs = random() % (10*1000000);
    printf( "Sleeping %d usecs\n", usecs );
    usleep(usecs);
    printf( "Killing!\n" );
    kill( pid, SIGKILL );
    pthread_exit(NULL);
  }

  return NULL;

}

void run() {

  pthread_attr_t attr;
  pthread_attr_init(&attr);
  // set stack size to 32K, so we don't run out of memory
  pthread_attr_setstacksize(&attr, 32*1024);

  const int num_clients = 10;
  pthread_t ths[num_clients];
  for( int i = 0; i < num_clients; i++ ) {
    assert( pthread_create( &ths[i], &attr, do_client, NULL ) == 0 );
  }
  
  for( int i = 0; i < num_clients; i++ ) {
    assert( pthread_join( ths[i], NULL ) == 0 );
  }

  pthread_attr_destroy(&attr);

}

int main(int argc, char *argv[])
{
  const char *username = "strib";
  const char *key_dir = "/home/strib/projects/dfs/wfs/auth";

  setlinebuf(stdout);

  // debugging 
  WFSLogInit _loginit( "none", "/none", 100, 5 );
  //wfs_set_debug(2);

  srandom(getpid());
  int port = htons(20000 + (getpid() % 10000));
  
  // start the server. automatically starts a thread
  // to listen on the UDP port and dispatch calls,
  // each in its own thread. you probably only need
  // one rpcs per process, though it doesn't hurt
  // to have more.
  rpc_enable_security(wfs_public_key_identifies_user,
		      wfs_get_private_key,
		      username);
  wfs_set_key_dir(key_dir);
  rpcs *server = new rpcs(port);

  // register a few server RPC handlers.
  srv service;
  server->reg(25, &service, &srv::handle_big);

  sleep(2);

  // server's address.
  bzero(&dst, sizeof(dst));
  dst.sin_family = AF_INET;
  dst.sin_addr.s_addr = inet_addr("127.0.0.1");
  dst.sin_port = port;

  const int num_runs = 20;
  for( int i = 0; i < num_runs; i++ ) {
    printf( "%d: Starting run %d\n", getpid(), i );
    run();
  }

  sleep(2);

  delete server;
  server = NULL;

  exit(0);
}
